# Copyright 2019 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Functions and classes involved in training

Notes
-----
The loss functions :class:`HardOptimalCompletionDistillationLoss` and
:class:`MinimumErrorRateLoss` have been moved to :mod:`pydrobert.torch.layers`
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import math
import warnings

from csv import DictReader, writer
from string import Formatter
from collections import OrderedDict

import torch
import param


__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2019 Sean Robertson"
__all__ = [
    'TrainingStateParams',
    'TrainingStateController',
]


class TrainingStateParams(param.Parameterized):
    '''Parameters controlling a TrainingStateController

    This class implements the
    :class:`pydrobert.param.optuna.TunableParameterized` interface
    '''
    num_epochs = param.Integer(
        None, bounds=(1, None), softbounds=(10, 100),
        doc='Total number of epochs to run for. If unspecified, runs '
        'until the early stopping criterion (or infinitely if disabled) '
    )
    log10_learning_rate = param.Number(
        None, softbounds=(-10, -2),
        doc='Initial optimizer log-learning rate. If unspecified, the initial '
        'learning rate of the optimizer instance remains unchanged'
    )
    early_stopping_threshold = param.Number(
        0.0, bounds=(0, None), softbounds=(0, 1.),
        doc='Minimum magnitude decrease in validation metric from the last '
        'best that resets the early stopping clock. If zero, early stopping '
        'will never be performed'
    )
    early_stopping_patience = param.Integer(
        1, bounds=(1, None), softbounds=(1, 30),
        doc='Number of epochs after which, if the classifier has failed to '
        'decrease its validation metric by a threshold, training is '
        'halted'
    )
    early_stopping_burnin = param.Integer(
        0, bounds=(0, None), softbounds=(0, 10),
        doc='Number of epochs before the early stopping criterion kicks in'
    )
    reduce_lr_threshold = param.Number(
        0.0, bounds=(0, None), softbounds=(0, 1.),
        doc='Minimum magnitude decrease in validation metric from the last '
        'best that resets the clock for reducing the learning rate. If zero, '
        'the learning rate will never be reduced'
    )
    reduce_lr_factor = param.Magnitude(
        0.1, softbounds=(.1, .5), inclusive_bounds=(False, False),
        doc='Factor by which to multiply the learning rate if there has '
        'been no improvement in the  after "reduce_lr_patience" '
        'epochs'
    )
    reduce_lr_patience = param.Integer(
        1, bounds=(1, None), softbounds=(1, 30),
        doc='Number of epochs after which, if the classifier has failed to '
        'decrease its validation metric by a threshold, the learning rate is '
        'reduced'
    )
    reduce_lr_cooldown = param.Integer(
        0, bounds=(0, None), softbounds=(0, 10),
        doc='Number of epochs after reducing the learning rate before we '
        'resume checking improvements'
    )
    reduce_lr_log10_epsilon = param.Number(
        -8, bounds=(None, 0),
        doc='The log10 absolute difference between learning rates that, '
        'below which, reducing the learning rate is considered meaningless'
    )
    reduce_lr_burnin = param.Integer(
        0, bounds=(0, None), softbounds=(0, 10),
        doc='Number of epochs before the criterion for reducing the learning '
        'rate kicks in'
    )
    seed = param.Integer(
        None,
        doc='Seed used for training procedures (e.g. dropout). If '
        "unset, will not touch torch's seeding"
    )
    keep_last_and_best_only = param.Boolean(
        True,
        doc='If the model is being saved, keep only the model and optimizer '
        'parameters for the last and best epoch (in terms of validation loss).'
        ' If False, save every epoch. See also "saved_model_fmt" and '
        '"saved_optimizer_fmt"'
    )
    saved_model_fmt = param.String(
        'model_{epoch:03d}.pt',
        doc='The file name format string used to save model state information.'
        ' Entries from the state csv are used to format this string (see '
        'TrainingStateController)'
    )
    saved_optimizer_fmt = param.String(
        'optim_{epoch:03d}.pt',
        doc='The file name format string used to save optimizer state '
        'information. Entries from the state csv are used to format this '
        'string (see TrainingStateController)'
    )

    @classmethod
    def get_tunable(cls):
        '''Returns a set of tunable parameters'''
        return {
            'num_epochs', 'log10_learning_rate', 'early_stopping_threshold',
            'early_stopping_patience', 'early_stopping_burnin',
            'reduce_lr_factor', 'reduce_lr_threshold',
            'reduce_lr_patience', 'reduce_lr_cooldown',
            'reduce_lr_burnin',
        }

    @classmethod
    def suggest_params(cls, trial, base=None, only=None, prefix=''):
        '''Populate a parameterized instance with values from trial'''
        if only is None:
            only = cls.get_tunable()
        params = cls() if base is None else base
        pdict = params.param.params()
        if 'log10_learning_rate' in only:
            softbounds = pdict['log10_learning_rate'].get_soft_bounds()
            params.log10_learning_rate = trial.suggest_uniform(
                prefix + 'log10_learning_rate', *softbounds)
        if 'num_epochs' in only:
            softbounds = pdict['num_epochs'].get_soft_bounds()
            params.num_epochs = trial.suggest_int(
                prefix + 'num_epochs', *softbounds)
        if params.num_epochs is None:
            num_epochs = float('inf')
        else:
            num_epochs = params.num_epochs
        # if we sample patience and burnin so that their collective total
        # reaches or exceeds the number of epochs, they are effectively
        # disabled. Rather than allowing vast sums above the number of epochs,
        # we only allow the sum to reach the remaining epochs
        remaining_epochs = num_epochs
        if 'early_stopping_patience' not in only:
            remaining_epochs -= params.early_stopping_patience
        if 'early_stopping_burnin' not in only:
            remaining_epochs -= params.early_stopping_burnin
        remaining_epochs = max(0, remaining_epochs)
        if remaining_epochs and 'early_stopping_threshold' in only:
            softbounds = pdict['early_stopping_threshold'].get_soft_bounds()
            params.early_stopping_threshold = trial.suggest_uniform(
                prefix + 'early_stopping_threshold', *softbounds)
        if not params.early_stopping_threshold:
            remaining_epochs = 0
        if remaining_epochs and 'early_stopping_patience' in only:
            softbounds = pdict['early_stopping_patience'].get_soft_bounds()
            softbounds = tuple(min(x, remaining_epochs) for x in softbounds)
            params.early_stopping_patience = trial.suggest_int(
                prefix + 'early_stopping_patience', *softbounds)
            remaining_epochs -= params.early_stopping_patience
            assert remaining_epochs >= 0
        if remaining_epochs and 'early_stopping_burnin' in only:
            softbounds = pdict['early_stopping_burnin'].get_soft_bounds()
            softbounds = tuple(min(x, remaining_epochs) for x in softbounds)
            params.early_stopping_burnin = trial.suggest_int(
                prefix + 'early_stopping_burnin', *softbounds)
            remaining_epochs -= params.early_stopping_burnin
            assert remaining_epochs >= 0
        # we do the same thing, but for the learning rate scheduler
        remaining_epochs = num_epochs
        if 'reduce_lr_patience' not in only:
            remaining_epochs -= params.reduce_lr_patience
        if 'reduce_lr_burnin' not in only:
            remaining_epochs -= params.reduce_lr_burnin
        remaining_epochs = max(0, remaining_epochs)
        if remaining_epochs and 'reduce_lr_threshold' in only:
            softbounds = pdict['reduce_lr_threshold'].get_soft_bounds()
            params.reduce_lr_threshold = trial.suggest_uniform(
                prefix + 'reduce_lr_threshold', *softbounds)
        if not params.reduce_lr_threshold:
            remaining_epochs = 0
        if remaining_epochs and 'reduce_lr_patience' in only:
            softbounds = pdict['reduce_lr_patience'].get_soft_bounds()
            softbounds = tuple(min(x, remaining_epochs) for x in softbounds)
            params.reduce_lr_patience = trial.suggest_int(
                prefix + 'reduce_lr_patience', *softbounds)
            remaining_epochs -= params.reduce_lr_patience
        if remaining_epochs and 'reduce_lr_burnin' in only:
            softbounds = pdict['reduce_lr_burnin'].get_soft_bounds()
            softbounds = tuple(min(x, remaining_epochs) for x in softbounds)
            params.reduce_lr_burnin = trial.suggest_int(
                prefix + 'reduce_lr_burnin', *softbounds)
        if remaining_epochs and 'reduce_lr_factor' in only:
            softbounds = pdict['reduce_lr_factor'].get_soft_bounds()
            params.reduce_lr_factor = trial.suggest_uniform(
                prefix + 'reduce_lr_factor', *softbounds)
        if remaining_epochs and 'reduce_lr_cooldown' in only:
            softbounds = pdict['reduce_lr_cooldown'].get_soft_bounds()
            params.reduce_lr_cooldown = trial.suggest_int(
                prefix + 'reduce_lr_cooldown', *softbounds)
        return params


class TrainingStateController(object):
    '''Controls the state of training a model

    This class is used to help both control and persist experiment information
    like the current epoch, the model parameters, and model error. It assumes
    that the values stored in `params` have not changed when resuming a run.
    It is also used to control learning rates and early stopping.

    Examples
    --------

    >>> params = TrainingStateParams(num_epochs=5)
    >>> model = torch.nn.Linear(10, 1)
    >>> optimizer = torch.optim.Adam(model.parameters())
    >>> controller = TrainingStateController(
    ...    params,
    ...    state_csv_path='log.csv',
    ...    state_dir='states')
    >>> # load previous
    >>> controller.load_model_and_optimizer_for_epoch(model, optimizer)
    >>> for epoch in range(params.num_epochs):
    >>>     # do training loop for epoch
    >>>     train_loss, val_loss = 0.1, 0.01
    >>>     if not controller.update_for_epoch(
    ...             model, optimizer, train_loss, val_loss):
    >>>         break  # early stopping

    Parameters
    ----------
    params : TrainingStateParams
    state_csv_path : str, optional
        A path to where training state information is stored. It stores in
        comma-separated-values format the following information. Note that
        stored values represent the state *after* updates due to epoch
        results, such as the learning rate. That way, an experiment can be
        resumed without worrying about updating the loaded results

        1. "epoch": the epoch associated with this row of information
        2. "es_resume_cd": the number of epochs left before the early
           stopping criterion begins/resumes
        3. es_patience_cd: the number of epochs left that must pass
           without much improvement before training halts due to early stopping
        4. "rlr_resume_cd": the number of epochs left before the
           criterion for reducing the learning rate begins/resumes
        5. "rlr_patience_cd": the number of epochs left that must pass
           without much improvement before the learning rate is reduced
        6. "lr": the learning rate of the optimizer after any updates
        7. "train_met": mean training metric in exponent format. The metric
           is assumed to be lower is better
        8. "val_met": mean validation metric in exponent format. The metric
           is assumed to be lower is better
        9. Any additional entries added through :func:`add_entry`

        If unset, the history will not be stored/loaded
    state_dir : str, optional
        A path to a directory to store/load model and optimizer states. If
        unset, the information will not be stored/loaded
    warn : bool, optional
        Whether to warn using :mod:`warnings` module when a format string does
        not contain the "epoch" field

    Attributes
    ----------
    params : TrainingStateParams
    state_csv_path : str or None
    state_dir : str or None
    user_entry_types : OrderedDict
        A collection of user entries specified by :func:`add_entry`
    cache_hist : dict
        A dictionary of cached results per epoch. Is not guaranteed to be
        up-to-date with `state_csv_path` unless :func:`update_cache` is called
    fmt_dict : dict
        A dictionary of format strings for the CSV entries
    '''

    def __init__(self, params, state_csv_path=None, state_dir=None, warn=True):
        super(TrainingStateController, self).__init__()
        self.params = params
        if warn:
            for s in (
                    self.params.saved_model_fmt,
                    self.params.saved_optimizer_fmt):
                if not any(x[1] == 'epoch' for x in Formatter().parse(s)):
                    warnings.warn(
                        'State format string "{}" does not contain "epoch" '
                        'field, so is possibly not unique. In this case, only '
                        'the state of the last epoch will persist. To '
                        'suppress this warning, set warn=False'.format(s))
        self.state_csv_path = state_csv_path
        self.state_dir = state_dir
        self.cache_hist = dict()
        self.user_entry_types = OrderedDict()
        self.fmt_dict = dict()
        if params.num_epochs is None:
            self.fmt_dict['epoch'] = '{:010d}'
        else:
            self.fmt_dict['epoch'] = '{{:0{}d}}'.format(
                int(math.log10(params.num_epochs)) + 1)
        self.fmt_dict['es_resume_cd'] = '{{:0{}d}}'.format(
            int(math.log10(max(params.early_stopping_burnin, 1))) + 1)
        self.fmt_dict['es_patience_cd'] = '{{:0{}d}}'.format(
            int(math.log10(max(params.early_stopping_patience, 1))) + 1)
        self.fmt_dict['rlr_resume_cd'] = '{{:0{}d}}'.format(
            int(math.log10(max(
                params.reduce_lr_cooldown,
                params.reduce_lr_burnin,
                1,
            ))) + 1
        )
        self.fmt_dict['rlr_patience_cd'] = '{{:0{}d}}'.format(
            int(math.log10(max(params.reduce_lr_patience, 1))) + 1)
        self.fmt_dict['lr'] = '{{:.{}e}}'.format(self.SCIENTIFIC_PRECISION - 1)
        self.fmt_dict['train_met'] = self.fmt_dict['lr']
        self.fmt_dict['val_met'] = self.fmt_dict['lr']

    '''The number of digits in significand of scientific notation

    Controls how many digits are saved when writing metrics and learning rate
    to disk (i.e. the ``x`` in ``x * 10^y``). Used when generating the format
    strings in ``self.fmt_dict`` on initialization
    '''
    SCIENTIFIC_PRECISION = 5

    def add_entry(self, name, type_=str, fmt='{}'):
        '''Add an entry to to be stored and retrieved at every epoch

        This method is useful when training loops need specialized, persistent
        information on every epoch. Prior to the first time any information is
        saved via :func:`update_for_epoch`, this method can be called with an
        entry `name` and optional `type_`. The user is then expected to provide
        a keyword argument with that `name` every time :func:`update_for_epoch`
        is called. The values of those entries can be retrieved via
        :func:`get_info`, cast to `type_`, for any saved epoch

        Parameters
        ----------
        name : str
        type_ : type, optional
            `type_` should be a type that is serialized from a string via
            ``type_(str_obj)`` and serialized to a string via
            ``fmt.format(obj)``
        fmt : str, optional
            The format string used to serialize the objects into strings

        Examples
        --------
        >>> params = TrainingStateParams()
        >>> controller = TrainingStateController(params)
        >>> model = torch.nn.Linear(10, 1)
        >>> optimizer = torch.optim.Adam(model.parameters())
        >>> controller.add_entry('important_value', int)
        >>> controller.update_for_epoch(
        ...     model, optimizer, 0.1, 0.1, important_value=3)
        >>> controller.update_for_epoch(
        ...     model, optimizer, 0.2, 0.01, important_value=4)
        >>> assert controller[1]['important_value'] == 3
        >>> assert controller[2]['important_value'] == 4

        Notes
        -----
        :func:`add_entry` must be called prior to :func:`update_for_epoch`
        or :func:`save_info_to_hist`, or it may corrupt the experiment history.
        However, the controller can safely ignore additional entries when
        loading history from a CSV. Thus, there is no need to call
        :func:`add_entry` if no new training is to be done (unless those
        entries are needed outside of training)
        '''
        if name in {
                "epoch", "es_resume_cd", "es_patience_cd", "rlr_resume_cd",
                "rlr_patience_cd", "lr", "train_met", "val_met"}:
            raise ValueError('"{}" is a reserved entry name'.format(name))
        if not isinstance(type_, type):
            raise ValueError('type_ ({}) must be a type'.format(type_))
        self.user_entry_types[name] = type_
        self.fmt_dict[name] = fmt
        self.update_cache()

    def update_cache(self):
        '''Update the cache with history stored in state_csv_path'''
        # add a dummy entry for epoch "0" just to make logic easier. We
        # won't save it
        self.cache_hist[0] = {
            'epoch': 0,
            'es_resume_cd': self.params.early_stopping_burnin,
            'es_patience_cd': self.params.early_stopping_patience,
            'rlr_resume_cd': self.params.reduce_lr_burnin,
            'rlr_patience_cd': self.params.reduce_lr_patience,
            'train_met': float('inf'),
            'val_met': float('inf'),
            'lr': None,
        }
        self.cache_hist[0].update(
            dict((key, None) for key in self.user_entry_types))
        if self.params.log10_learning_rate is not None:
            self.cache_hist[0]['lr'] = (
                10 ** self.params.log10_learning_rate)
        if (self.state_csv_path is None or
                not os.path.exists(self.state_csv_path)):
            return
        with open(self.state_csv_path) as f:
            reader = DictReader(f)
            for row in reader:
                epoch = int(row['epoch'])
                self.cache_hist[epoch] = {
                    'epoch': epoch,
                    'es_resume_cd': int(row['es_resume_cd']),
                    'es_patience_cd': int(row['es_patience_cd']),
                    'rlr_resume_cd': int(row['rlr_resume_cd']),
                    'rlr_patience_cd': int(row['rlr_patience_cd']),
                    'lr': float(row['lr']),
                    'train_met': float(row['train_met']),
                    'val_met': float(row['val_met']),
                }
                for name, type_ in self.user_entry_types.items():
                    self.cache_hist[epoch][name] = type_(row[name])

    def get_last_epoch(self):
        '''int : last finished epoch from training, or 0 if no history'''
        self.update_cache()
        return max(self.cache_hist)

    def get_best_epoch(self, train_met=False):
        '''Get the epoch that has lead to the best validation metric val so far

        The "best" is the lowest recorded validation metric. In the case of
        ties, the earlier epoch is chosen.

        Parameters
        ----------
        train_met : bool, optional
            If :obj:`True` look for the best training metric value instead

        Returns
        -------
        epoch : int
            The corresponding 'best' epoch, or :obj:`0` if no epochs have run

        Notes
        -----
        Negligible differences between epochs are determined by
        :obj:`TrainingStateController.METRIC_PRECISION`, which is relative
        to the metrics base 10. This is in contrast to early stopping criteria
        and learning rate annealing, whose thresholds are absolute.
        '''
        ent = 'train_met' if train_met else 'val_met'
        fmt = self.fmt_dict[ent]
        self.update_cache()
        min_epoch = 0
        min_met = self.cache_hist[0][ent]
        min_met = float(fmt.format(min_met))
        for info in self.cache_hist.values():
            cur = float(fmt.format(info[ent]))
            if cur < min_met:
                min_epoch = info['epoch']
                min_met = cur
        return min_epoch

    def load_model_for_epoch(self, model, epoch=None, strict=True):
        '''Load up just the model, or initialize it

        Parameters
        ----------
        model : torch.nn.Module
            Model state will be loaded into this
        epoch : int or :obj:`None`, optional
            The epoch from which the states should be loaded. We look for the
            appropriately named files in ``self.state_dir``. If `epoch` is
            :obj:`None`, the best epoch in recorded history will be loaded. If
            it's 0, the model is initialized with states from the beginning of
            the experiment
        strict : bool, optional
            Whether to strictly enforce that the keys in ``model.state_dict()``
            match those that were saved
        '''
        if epoch is None:
            epoch = self.get_best_epoch()
        if not epoch:
            if self.params.seed is not None:
                torch.manual_seed(self.params.seed)
            if hasattr(model, 'reset_parameters'):
                model.reset_parameters()
            else:
                warnings.warn(
                    'model has no reset_parameters() method, so cannot '
                    'reset parameters for epoch 0'
                )
        elif self.state_dir is not None:
            epoch_info = self[epoch]
            model_basename = self.params.saved_model_fmt.format(**epoch_info)
            model_state_dict = torch.load(
                os.path.join(self.state_dir, model_basename),
                map_location='cpu',
            )
            model.load_state_dict(model_state_dict, strict=strict)
        else:
            warnings.warn(
                'Unable to load model for epoch {}. No state directory!'
                ''.format(epoch)
            )

    def load_model_and_optimizer_for_epoch(
            self, model, optimizer, epoch=None, strict=True):
        '''Load up model and optimizer states, or initialize them

        Parameters
        ----------
        model : torch.nn.Module
            Model state will be loaded into this
        optimizer : torch.optim.Optimizer
            Optimizer state will be loaded into this
        epoch : int or :obj:`None`, optional
            The epoch from which the states should be loaded. We look for the
            appropriately named files in ``self.state_dir``. If `epoch` is
            :obj:`None`, the last epoch in recorded history will be loaded. If
            it's 0, the model and optimizer are initialized with states
            for the beginning of the experiment.
        strict : bool, optional
            Whether to strictly enforce that the keys in ``model.state_dict()``
            match those that were saved
        '''
        if epoch is None:
            epoch = self.get_last_epoch()
        if not epoch:
            if self.params.seed is not None:
                torch.manual_seed(self.params.seed)
            if hasattr(model, 'reset_parameters'):
                model.reset_parameters()
            else:
                warnings.warn(
                    'model has no reset_parameters() method, so cannot '
                    'reset parameters for epoch 0'
                )
            if self.params.log10_learning_rate is not None:
                for param_group in optimizer.param_groups:
                    param_group['lr'] = 10 ** self.params.log10_learning_rate
            # there is no public API for resetting the state dictionary, so
            # we create a new instance as best as possible and copy the state
            # over from there. Note that settings like weight decay are already
            # part of the parameter group, so we don't need to worry about
            # initializing with them.
            brand_new_optimizer = type(optimizer)(optimizer.param_groups)
            optimizer.load_state_dict(brand_new_optimizer.state_dict())
            del brand_new_optimizer
        elif self.state_dir is not None:
            epoch_info = self[epoch]
            model_basename = self.params.saved_model_fmt.format(**epoch_info)
            optimizer_basename = self.params.saved_optimizer_fmt.format(
                **epoch_info)
            model_state_dict = torch.load(
                os.path.join(self.state_dir, model_basename),
                map_location='cpu',
            )
            model.load_state_dict(model_state_dict, strict=strict)
            optimizer_state_dict = torch.load(
                os.path.join(self.state_dir, optimizer_basename),
                map_location='cpu',
            )
            optimizer.load_state_dict(optimizer_state_dict)
        else:
            warnings.warn(
                'Unable to load model and optimizer for epoch {}. No state'
                'directory!'.format(epoch)
            )

    def delete_model_and_optimizer_for_epoch(self, epoch):
        '''Delete state dicts for model and epoch off of disk, if they exist

        This method does nothing if the epoch records or the files do not
        exist. It is called during :func:`update_for_epoch` if the parameter
        ``keep_last_and_best_only`` is :obj:`True`

        Parameters
        ----------
        epoch : int
            The epoch in question
        '''
        if self.state_dir is None:
            return
        epoch_info = self.get_info(epoch, None)
        if epoch_info is None:
            return
        model_basename = self.params.saved_model_fmt.format(**epoch_info)
        optimizer_basename = self.params.saved_optimizer_fmt.format(
            **epoch_info)
        try:
            os.remove(os.path.join(self.state_dir, model_basename))
        except OSError:
            pass
        try:
            os.remove(os.path.join(self.state_dir, optimizer_basename))
        except OSError:
            pass

    def get_info(self, epoch, *default):
        '''Get history entries for a specific epoch

        If there's an entry present for `epoch`, return it. The value is a
        dictionary with the keys "epoch", "es_resume_cd", "es_patience_cd",
        "rlr_resume_cd", "rlr_patience_cd", "lr", "train_met", and "val_met",
        as well as any additional entries specified through :func:`add_entry`.

        If there's no entry for `epoch`, and no additional arguments were
        passed to this method, it raises a :class:`KeyError`. If an additional
        argument was passed to this method, return it.
        '''
        if len(default) > 1:
            raise TypeError('expected at most 2 arguments, got 3')
        if epoch in self.cache_hist:
            return self.cache_hist[epoch]
        self.update_cache()
        return self.cache_hist.get(epoch, *default)

    def __getitem__(self, epoch):
        return self.get_info(epoch)

    def save_model_and_optimizer_with_info(self, model, optimizer, info):
        '''Save model and optimizer state dictionaries to file given epoch info

        This is called automatically during :func:`update_for_epoch`. Does not
        save if there is no directory to save to (i.e. ``self.state_dir is
        None``). Format strings from ``self.params`` are formatted with the
        values from `info` to construct the base names of each file

        Parameters
        ----------
        model : AcousticModel
        optimizer : torch.optim.Optimizer
        info : dict
            A dictionary with the entries "epoch", "es_resume_cd",
            "es_patience_cd", "rlr_resume_cd", "rlr_patience_cd", "lr",
            "train_met", "val_met", and any entries specified through
            :func:`add_entry`
        '''
        if self.state_dir is None:
            return
        if not os.path.isdir(self.state_dir):
            os.makedirs(self.state_dir)
        model_basename = self.params.saved_model_fmt.format(**info)
        optimizer_basename = self.params.saved_optimizer_fmt.format(
            **info)
        torch.save(
            model.state_dict(),
            os.path.join(self.state_dir, model_basename),
        )
        torch.save(
            optimizer.state_dict(),
            os.path.join(self.state_dir, optimizer_basename),
        )

    def save_info_to_hist(self, info):
        '''Append history entries to the history csv

        This is called automatically during :func:`update_for_epoch`. Does not
        save if there is no file to save to (i.e. ``self.state_csv_path is
        None``). Values are appended to the end of the csv file - no checking
        is performed for mantaining a valid history.

        Parameters
        ----------
        info : dict
            A dictionary with the entries "epoch", "es_resume_cd",
            "es_patience_cd", "rlr_resume_cd", "rlr_patience_cd", "lr",
            "train_met", "val_met", and any other entries specified via
            :func:`add_entry`
        '''
        self.cache_hist[info['epoch']] = info
        if self.state_csv_path is None:
            return
        names = [
            'epoch',
            'es_resume_cd',
            'es_patience_cd',
            'rlr_resume_cd',
            'rlr_patience_cd',
            'lr',
            'train_met',
            'val_met',
        ]
        names += list(self.user_entry_types)
        write_header = not os.path.exists(self.state_csv_path)
        with open(self.state_csv_path, 'a') as f:
            wr = writer(f)
            if write_header:
                wr.writerow(names)
            wr.writerow([
                self.fmt_dict[k].format(info[k]) for k in names])

    def continue_training(self, epoch=None):
        '''Return a boolean on whether to continue training

        Useful when resuming training. Will check the training history at the
        target `epoch` and determine whether training should continue from that
        point, based on the total number of epochs and the early stopping
        criterion.

        Parameters
        ----------
        epoch : int or :obj:`None`, optional
            The epoch to check the history of. If :obj:`None`, the last epoch
            will be inferred

        Returns
        -------
        cont : bool
            :obj:`True` if training should continue
        '''
        if epoch is None:
            epoch = self.get_last_epoch()
        info = self[epoch]
        if not self.params.num_epochs:
            cont = True
        else:
            cont = epoch < self.params.num_epochs
        if self.params.early_stopping_threshold and not info["es_patience_cd"]:
            cont = False
        return cont

    def update_for_epoch(
            self, model, optimizer, train_met, val_met, epoch=None,
            **kwargs):
        '''Update history and optimizer after latest epoch results

        Parameters
        ----------
        model : AcousticModel
        optimizer : torch.optim.Optimizer
        train_met : float
            Mean value of metric on training set for epoch
        val_met : float
            Mean value of metric on validation set for epoch
        epoch : int, optional
            The epoch that just finished. If unset, it is inferred to be one
            after the last epoch in the history
        kwargs : Keyword arguments, optional
            Additional keyword arguments can be used to specify the values
            of entries specified via :func:`add_entry`

        Returns
        -------
        cont : bool
            Whether to continue training. This can be set to :obj:`False`
            either by hitting the max number of epochs or by early stopping
        '''
        if epoch is None:
            epoch = self.get_last_epoch() + 1
        if not self.params.num_epochs:
            cont = True
        else:
            cont = epoch < self.params.num_epochs
            if epoch > self.params.num_epochs:
                warnings.warn(
                    'Training is continuing, despite passing num_epochs')
        info = dict(self.get_info(epoch - 1, None))
        if info is None:
            raise ValueError(
                'no entry for the previous epoch {}, so unable to update'
                ''.format(epoch))
        for key, value in kwargs.items():
            if key not in self.user_entry_types:
                raise TypeError(
                    'update_for_epoch() got an unexpected keyword argument '
                    '"{}" (did you forget to add_entry()?)'.format(key))
            elif not isinstance(value, self.user_entry_types[key]):
                raise ValueError(
                    'keyword argument "{}" value is not of type {}'
                    ''.format(key, self.user_entry_types[key]))
            info[key] = value
        remaining_user_entries = set(self.user_entry_types) - set(kwargs)
        if remaining_user_entries:
            raise TypeError(
                'The following keyword arguments were not provided as keyword '
                'arguments but were specified via add_entry(): {}'
                ''.format(sorted(remaining_user_entries)))
        last_best = self.get_best_epoch()
        best_info = self[last_best]
        if info['lr'] is None:
            # can only happen during the first epoch. We don't know the
            # optimizer defaults, so we get them now
            info['lr'] = optimizer.defaults['lr']
        if info["es_resume_cd"]:
            info["es_resume_cd"] -= 1
        elif (max(best_info['val_met'] - val_met, 0) <
                self.params.early_stopping_threshold):
            info["es_patience_cd"] -= 1
            if info["es_patience_cd"] < 0:
                warnings.warn(
                    "Early stopping criterion was already met, but training "
                    "has continued")
                info["es_patience_cd"] = 0
        else:
            info["es_patience_cd"] = self.params.early_stopping_patience
        # we do it this way in case someone continues training after early
        # stopping has been reached
        if self.params.early_stopping_threshold and not info["es_patience_cd"]:
            cont = False
        if info["rlr_resume_cd"]:
            info["rlr_resume_cd"] -= 1
        elif (max(best_info['val_met'] - val_met, 0) <
                self.params.reduce_lr_threshold):
            info["rlr_patience_cd"] -= 1
            if not info["rlr_patience_cd"]:
                old_lr = info['lr']
                new_lr = old_lr * self.params.reduce_lr_factor
                rlr_epsilon = 10 ** self.params.reduce_lr_log10_epsilon
                if old_lr - new_lr > rlr_epsilon:
                    info['lr'] = new_lr
                    for param_group in optimizer.param_groups:
                        # just assume that the user knows what's what if
                        # the optimizer's lr doesn't match the old one
                        param_group['lr'] = new_lr
                info["rlr_resume_cd"] = self.params.reduce_lr_cooldown
                info["rlr_patience_cd"] = self.params.reduce_lr_patience
        else:
            info["rlr_patience_cd"] = self.params.reduce_lr_patience
        info["epoch"] = epoch
        info["val_met"] = val_met
        info["train_met"] = train_met
        # in the unlikely event that there's a SIGTERM here, this block tries
        # its best to maintain a valid history on exit. We have to delete the
        # old states first in case the file names match the new states
        self.cache_hist[info['epoch']] = info
        cur_best = self.get_best_epoch()
        try:
            if self.params.keep_last_and_best_only and cur_best != epoch - 1:
                self.delete_model_and_optimizer_for_epoch(epoch - 1)
            if self.params.keep_last_and_best_only and cur_best != last_best:
                self.delete_model_and_optimizer_for_epoch(last_best)
        finally:
            self.save_model_and_optimizer_with_info(model, optimizer, info)
            self.save_info_to_hist(info)
        return cont



#! /usr/bin/env python

import os
import sys
import pandas as pd
import matplotlib.pyplot as plt
import torch

# remove this line if you've installed pydrobert-torch
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.realpath(__file__)))))

from pydrobert.torch.estimators import *
from pydrobert.torch.util import *

eos, padding = 0, -1
batch_size, inp_size, num_classes, num_samps = 10, 20, 5, 100
T, S, sos = 30, 10, -1
hidden_size = 40
num_iters, num_seeds = 1000, 10
dict_ = dict()
c_ff = c_rnn = None


def f(hyp, ref, gamma=.95):
    dists = prefix_error_rates(
        ref, hyp.long(), eos=eos, norm=False, padding=-1)
    r = -(dists[1:] - dists[:-1])
    r = r.masked_fill(dists[1:].eq(padding), 0.)
    R = time_distributed_return(r, gamma)
    return R


def c(z, lens):
    # you can avoid this sorting business by using enforce_sorted=False in
    # later versions of pytorch
    lens, idxs = lens.sort(descending=True)
    _, rev_idxs = idxs.sort()
    z = z.index_select(1, idxs)
    zp = torch.nn.utils.rnn.pack_padded_sequence(z, lens)
    zp, _ = c_rnn(zp)
    zp = torch.nn.utils.rnn.PackedSequence(c_ff(zp[0]), zp[1])
    z, _ = torch.nn.utils.rnn.pad_packed_sequence(zp, total_length=z.shape[0])
    z = z.squeeze(-1)
    z = z.index_select(1, rev_idxs)
    return z


# estimator format: '<estimator_name>[-{d,g}-<hs>]'
# where <estimator_name> is 'reinforce' or 'relax'
# <hs> is the hidden size of the control variate
# "d" means use difference as control variate loss (i.e. MSE)
# "g" means use variance of entire gradient estimate as control variate loss
estimators = (
    'reinforce',
    'reinforce-d-5',
    'reinforce-d-10',
    'reinforce-g-10',
    'reinforce-d-20',
    'relax-g-5',
    'relax-d-10',
    'relax-g-10',
    'relax-g-20',
)

for estimator in estimators:
    if len(sys.argv) > 1:
        continue
    if '-' in estimator:
        cv_hidden_size = int(estimator.split('-')[-1])
    else:
        cv_hidden_size = 1
    for seed in range(num_seeds):
        torch.manual_seed(seed)
        inp = torch.randn(T, batch_size, inp_size)
        ref_lens = torch.randint(1, S + 1, (batch_size,))
        ref = torch.nn.utils.rnn.pad_sequence(
            [torch.randint(1, num_classes, (x + 1,)) for x in ref_lens],
            padding_value=padding,
        )
        ref[ref_lens, range(batch_size)] = eos
        # repeat the same reference transcription for each sample
        ref_rep = ref.unsqueeze(-1).repeat(1, 1, num_samps)
        cell = torch.nn.RNNCell(inp_size + 1, hidden_size)
        ff = torch.nn.Linear(hidden_size, num_classes)
        c_rnn = torch.nn.RNN(num_classes, cv_hidden_size, bidirectional=True)
        c_ff = torch.nn.Linear(2 * cv_hidden_size, 1)
        optim = torch.optim.Adam(
            tuple(cell.parameters()) + tuple(ff.parameters()) +
            tuple(c_rnn.parameters()) + tuple(c_ff.parameters())
        )
        ers = []
        c_losses = []
        c_loss = float('nan')
        for t in range(num_iters):
            h_t = torch.zeros(batch_size, 1, hidden_size)
            hyp = torch.full((1, batch_size, 1), sos, dtype=torch.long)
            optim.zero_grad()
            logits = z = None
            for inp_t in inp:
                hyp_tm1 = hyp[-1]
                old_samp = hyp_tm1.shape[-1]
                inp_t = inp_t.unsqueeze(1).expand(
                    batch_size, old_samp, inp_size)
                x_t = torch.cat([inp_t, hyp_tm1.unsqueeze(2).float()], -1)
                h_t = cell(
                    x_t.view(batch_size * old_samp, inp_size + 1),
                    h_t.view(batch_size * old_samp, hidden_size),
                ).view(batch_size, old_samp, hidden_size)
                logits_t = ff(h_t)  # (batch_size, old_samp, num_classes)
                hyp, z_t = random_walk_advance(
                    logits_t, num_samps, hyp, eos, include_relaxation=True)
                if old_samp == 1:
                    h_t = h_t.repeat(1, num_samps, 1).contiguous()
                    logits = logits_t.unsqueeze(0).expand(
                        -1, -1, num_samps, -1)
                    z = z_t.unsqueeze(0)
                else:
                    logits = torch.cat([logits, logits_t.unsqueeze(0)], dim=0)
                    z = torch.cat([z, z_t.unsqueeze(0)], dim=0)
            ref_rep = ref_rep.view(-1, batch_size * num_samps)
            hyp = hyp[1:].view(-1, batch_size * num_samps)  # get rid of sos
            logits = logits.view(-1, batch_size * num_samps, num_classes)
            z = z.view(-1, batch_size * num_samps, num_classes)
            mask = torch.isinf(z).any(-1, keepdim=True)
            lens = mask.eq(0).long().sum(0).squeeze(-1)
            fb = f(hyp, ref_rep)
            if estimator.startswith('reinforce'):
                if '-' in estimator:
                    fb = diff = fb - c(logits.detach(), lens)
                g = reinforce(fb, hyp, logits, 'cat')
            else:
                (
                    diff, dlog_pb, dc_z, dc_z_tilde,
                ) = relax(
                    fb, hyp, logits, z, c, 'cat', lens=lens, components=True)
                g = diff * dlog_pb + dc_z - dc_z_tilde
            g = g.masked_fill(mask, 0.)
            logits.backward(-g)
            if g.grad_fn is not None:
                if '-g-' in estimator:
                    c_loss = (g ** 2).sum(0)
                else:
                    c_loss = (diff ** 2).sum(0)
                c_loss = (c_loss / lens.float().unsqueeze(-1)).mean()
                c_loss.backward()
                c_loss = c_loss.item()
            c_losses.append(c_loss)
            optim.step()
            ers.append(error_rate(ref_rep, hyp, eos=eos).mean().item())
            print(t, ers[-1], c_losses[-1])
        dict_[(estimator, 'ers', seed)] = pd.Series(ers)
        dict_[(estimator, 'c_losses', seed)] = pd.Series(c_losses)

if dict_:
    df = pd.DataFrame.from_dict(dict_, orient='index')
    df.to_csv(
        'estimator_convergence.csv', index_label=['estimator', 'val', 'seed'])

df = pd.read_csv(
    'estimator_convergence.csv', index_col=['estimator', 'val', 'seed'])

colours = (
    'tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple',
    'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan',
)
fig, er_ax = plt.subplots()
c_loss_ax = er_ax.twinx()
for color, est in zip(colours, estimators):
    df_est = df.xs(est)
    df_est_ers = df_est.xs('ers')
    mu_ers = df_est_ers.mean()
    std_ers = df_est_ers.std()
    er_ax.plot(mu_ers.index.values, mu_ers, color=color, label=est)
    er_ax.fill_between(
        mu_ers.index.values, mu_ers - std_ers, mu_ers + std_ers, color=color,
        alpha=.5 / len(estimators))
    df_est_c_losses = df_est.xs('c_losses')
    mu_c_losses = df_est_c_losses.mean()
    if not mu_c_losses.isna().any():
        c_loss_ax.plot(
            mu_c_losses.index.values, mu_c_losses, color=color,
            linestyle='--')

er_ax.legend()
ax = fig.get_axes()[0]
er_ax.set_xticks(range(0, len(mu_ers), len(mu_ers) // 10))
er_ax.set_xlabel('iterations')
er_ax.set_ylabel('error rate')
c_loss_ax.set_ylabel('control variate loss')
fig.savefig('estimator_convergence.png', dpi=1080)



# Copyright 2020 Sean Robertson
#
# Code for polyharmonic_spline is converted from tensorflow code
# https://github.com/tensorflow/addons/blob/v0.11.2/tensorflow_addons/image/interpolate_spline.py
# code for sparse_image_warp is derived from tensorflow code, though it's not identical
# https://github.com/tensorflow/addons/blob/v0.11.2/tensorflow_addons/image/sparse_image_warp.py
#
# Which are also Apache 2.0 Licensed:
#
# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Utility functions"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import warnings

import torch
import pydrobert.torch

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2019 Sean Robertson"
__all__ = [
    "beam_search_advance",
    "dense_image_warp",
    "error_rate",
    "optimal_completion",
    "parse_arpa_lm",
    "polyharmonic_spline",
    "prefix_error_rates",
    "random_walk_advance",
    "sequence_log_probs",
    "sparse_image_warp",
    "time_distributed_return",
]


def parse_arpa_lm(file_, token2id=None):
    r"""Parse an ARPA statistical language model

    An `ARPA language model <https://cmusphinx.github.io/wiki/arpaformat/>`__
    is an n-gram model with back-off probabilities. It is formatted as

    ::

        \data\
        ngram 1=<count>
        ngram 2=<count>
        ...
        ngram <N>=<count>

        \1-grams:
        <logp> <token[t]> <logb>
        <logp> <token[t]> <logb>
        ...

        \2-grams:
        <logp> <token[t-1]> <token[t]> <logb>
        ...

        \<N>-grams:
        <logp> <token[t-<N>+1]> ... <token[t]>
        ...

        \end\

    Parameters
    ----------
    file_ : str or file
        Either the path or a file pointer to the file
    token2id : dict, optional
        A dictionary whose keys are token strings and values are ids. If set,
        tokens will be replaced with ids on read

    Returns
    -------
    prob_list : list
        A list of the same length as there are orders of n-grams in the
        file (e.g. if the file contains up to tri-gram probabilities then
        `prob_list` will be of length 3). Each element is a dictionary whose
        key is the word sequence (earliest word first). For 1-grams, this is
        just the word. For n > 1, this is a tuple of words. Values are either
        a tuple of ``logp, logb`` of the log-probability and backoff
        log-probability, or, in the case of the highest-order n-grams that
        don't need a backoff, just the log probability
    """
    if isinstance(file_, str):
        with open(file_) as f:
            return parse_arpa_lm(f, token2id=token2id)
    line = ""
    for line in file_:
        if line.strip() == "\\data\\":
            break
    if line.strip() != "\\data\\":
        raise IOError("Could not find \\data\\ line. Is this an ARPA file?")
    ngram_counts = []
    count_pattern = re.compile(r"^ngram\s+(\d+)\s*=\s*(\d+)$")
    for line in file_:
        line = line.strip()
        if not line:
            continue
        match = count_pattern.match(line)
        if match is None:
            break
        n, count = (int(x) for x in match.groups())
        if len(ngram_counts) < n:
            ngram_counts.extend(0 for _ in range(n - len(ngram_counts)))
        ngram_counts[n - 1] = count
    prob_list = [dict() for _ in ngram_counts]
    ngram_header_pattern = re.compile(r"^\\(\d+)-grams:$")
    ngram_entry_pattern = re.compile(r"^(-?\d+(?:\.\d+)?)\s+(.*)$")
    while line != "\\end\\":
        match = ngram_header_pattern.match(line)
        if match is None:
            raise IOError('line "{}" is not valid'.format(line))
        ngram = int(match.group(1))
        if ngram > len(ngram_counts):
            raise IOError(
                "{}-grams count was not listed, but found entry" "".format(ngram)
            )
        dict_ = prob_list[ngram - 1]
        for line in file_:
            line = line.strip()
            if not line:
                continue
            match = ngram_entry_pattern.match(line)
            if match is None:
                break
            logp, rest = match.groups()
            tokens = tuple(rest.strip().split())
            # IRSTLM and SRILM allow for implicit backoffs on non-final
            # n-grams, but final n-grams must not have backoffs
            logb = 0.0
            if len(tokens) == ngram + 1 and ngram < len(prob_list):
                try:
                    logb = float(tokens[-1])
                    tokens = tokens[:-1]
                except ValueError:
                    pass
            if len(tokens) != ngram:
                raise IOError(
                    'expected line "{}" to be a(n) {}-gram' "".format(line, ngram)
                )
            if token2id is not None:
                tokens = tuple(token2id[tok] for tok in tokens)
            if ngram == 1:
                tokens = tokens[0]
            if ngram != len(ngram_counts):
                dict_[tokens] = (float(logp), logb)
            else:
                dict_[tokens] = float(logp)
    if line != "\\end\\":
        raise IOError("Could not find \\end\\ line")
    for ngram_m1, (ngram_count, dict_) in enumerate(zip(ngram_counts, prob_list)):
        if len(dict_) != ngram_count:
            raise IOError(
                "Expected {} {}-grams, got {}".format(ngram_count, ngram_m1, len(dict_))
            )
    return prob_list


def beam_search_advance(
    logits_t,
    width,
    log_prior=None,
    y_prev=None,
    eos=pydrobert.torch.INDEX_PAD_VALUE,
    lens=None,
    prevent_eos=False,
    distribution=True,
):
    r"""Advance a beam search

    Suppose a model outputs a un-normalized log-probability distribution over
    the next element of a sequence in `logits_t` s.t.

    .. math::

        Pr(y_t = c ; log\_prior) = exp(logits_{t,c}) / \sum_k exp(logits_{t,k})

    We assume :math:`logits_t` is a function of what comes before
    :math:`logits_t = f(logits_{<t}, y_{<t})`. Alternatively, letting
    :math:`s_t = (logits_t, y_t)`, :math:`s` is a Markov Chain. A model is
    auto-regressive if :math:`f` depends on :math:`y_{<t}`, and is not
    auto-regressive if :math:`logits_t = f(logits_{<t})`.

    Beam search is a heuristic mechanism for determining a best path, i.e.
    :math:`\arg \max_y Pr(y)` that maximizes the probability of the best path
    by keeping track of `width` high probability paths called "beams" (the
    aggregate of which for a given batch element is named, unfortunately, "the
    beam"). If the model is auto-regressive, beam search is only approximate.
    However, if the model is not auto-regressive, beam search gives an exact
    n-best list.

    This function is called at every time step. It updates old beam
    log-probabilities (`log_prior`) with new ones (`score`) from the
    conditional derived from `logits_t`, and updates us the class indices
    emitted between them (`y`). See the examples section for how this might
    work.

    Parameters
    ----------
    logits_t : torch.FloatTensor
        The conditional probabilities over class labels for the current time
        step. Either of shape ``(batch_size, old_width, num_classes)``,
        where ``old_width`` is the number of beams in the previous time step,
        or ``(batch_size, num_classes)``, where it is assumed that
        ``old_width == 1``
    width : int
        The number of beams in the beam to produce for the current time step.
        ``width <= num_classes``
    log_prior : torch.FloatTensor, optional
        A tensor of (or proportional to) log prior probabilities of beams up
        to the previous time step. Either of shape ``(batch_size, old_width)``
        or ``(batch_size,)``. If unspecified, a uniform log prior will be used
    y_prev : torch.LongTensor, optional
        A tensor of shape ``(t - 1, batch_size, old_width)`` or
        ``(t - 1, batch_size)`` specifying :math:`y_{<t}`. If unspecified,
        it is assumed that ``t == 1``
    eos : int, optional
        A special end-of-sequence symbol indicating that the beam has ended.
        Can be a class index. If this value occurs in in ``y_prev[-1, bt, bm]``
        for some batch ``bt`` and beam ``bm``, that beam will be padded with
        an `eos` token and the score for that beam won't change
    lens : torch.LongTensor, optional
        A tensor of shape ``(batch_size,)``. If ``t > lens[bt]`` for some
        batch ``bt``, all beams for ``bt`` will be considered finished. All
        scores will be fixed and `eos` will be appended to `y_prev`
    prevent_eos : bool, optional
        Setting this flag to :obj:`True` will keep `eos` targets from entering
        a beam unless it has finished (either with a prior `eos` or through
        `lens`). Note that this will only have an effect when ``0 <= eos <=
        num_classes``
    distribution : bool, optional
        If :obj:`False`, a log-softmax will not be applied to `logits_t` prior
        to calculating the beams. Disabling `distribution` should be avoided
        when the beam search is being performed directly on model output.
        Setting `distribution` to :obj:`False` is useful when bootstrapping
        a language model probability distribution to the search

    Returns
    -------
    score : torch.FloatTensor
        Of shape ``(batch_size, width)`` of the log-joint probabilities of the
        new beams in the beam
    y : torch.LongTensor
        Of shape ``(t, batch_size, width)`` of indices of the class labels
        generated up to this point
    s : torch.LongTensor
        Of shape ``(batch_size, width)`` of indices of beams in the old beam
        which prefix the new beam. Note that beams in the new beam are sorted
        by descending probability

    Examples
    --------

    Auto-regressive decoding with beam search. We assume that all input have
    the same number of steps

    >>> N, I, C, T, W, H, eos, start = 5, 5, 10, 100, 5, 10, 0, -1
    >>> cell = torch.nn.RNNCell(I + 1, H)
    >>> ff = torch.nn.Linear(H, C)
    >>> inp = torch.rand(T, N, I)
    >>> y = torch.full((1, N, 1), start, dtype=torch.long)
    >>> h_t = torch.zeros(N, 1, H)
    >>> score = None
    >>> for inp_t in inp:
    >>>     y_tm1 = y[-1]
    >>>     old_width = y_tm1.shape[-1]
    >>>     inp_t = inp_t.unsqueeze(1).expand(N, old_width, I)
    >>>     x_t = torch.cat([inp_t, y_tm1.unsqueeze(2).float()], -1)
    >>>     h_t = cell(
    ...         x_t.view(N * old_width, I + 1),
    ...         h_t.view(N * old_width, H)
    ...     ).view(N, old_width, H)
    >>>     logits_t = ff(h_t)
    >>>     score, y, s_t = beam_search_advance(logits_t, W, score, y, eos)
    >>>     h_t = h_t.gather(1, s_t.unsqueeze(-1).expand(N, W, H))
    >>> bests = []
    >>> for batch_idx in range(N):
    >>>     best_beam_path = y[1:, batch_idx, 0]
    >>>     not_special_mask = best_beam_path.ne(eos)
    >>>     best_beam_path = best_beam_path.masked_select(not_special_mask)
    >>>     bests.append(best_beam_path)

    Were we to have a :class:`pydrobert.torch.layers.SequentialLanguageModel`,
    we could modify `logits_t`, and thus the search, to account for it:

    >>> # ... same as before
    >>> logits_t = torch.nn.functional.log_softmax(logits_t, -1)
    >>> logits_t = logits_t + lmb * lm(y)  # lmb is some constant
    >>> score, y, s_t = beam_search_advance(
    ...     logits_t, W, score, y, eos, distribution=False)
    >>> # ... same as before

    Note that `score` would no longer reflect a log-joint probability.

    The following produces a ``W``-best list for non-auto-regressive model. We
    don't emit an `eos`, instead completing the sequence when we've hit the
    target length via `lens`

    >>> N, I, C, T, W, H = 5, 5, 10, 100, 5, 10
    >>> rnn = torch.nn.RNN(I, H)
    >>> ff = torch.nn.Linear(H, C)
    >>> inp = torch.rand(T, N, I)
    >>> lens = torch.randint(1, T + 1, (N,)).sort(descending=True)[0]
    >>> packed_inp = torch.nn.utils.rnn.pack_padded_sequence(inp, lens)
    >>> packed_h, _ = rnn(packed_inp)
    >>> packed_logits = ff(packed_h[0])
    >>> logits = torch.nn.utils.rnn.pad_packed_sequence(
    ...     torch.nn.utils.rnn.PackedSequence(
    ...         packed_logits, batch_sizes=packed_h[1]),
    ...     total_length=T,
    ... )[0]
    >>> y = score = None
    >>> for t, logits_t in enumerate(logits):
    >>>     if t:
    >>>         logits_t = logits_t.unsqueeze(1).expand(-1, W, -1)
    >>>     score, y, _ = beam_search_advance(logits_t, W, score, y, lens=lens)
    """
    if logits_t.dim() == 2:
        logits_t = logits_t.unsqueeze(1)
    elif logits_t.dim() != 3:
        raise RuntimeError("logits_t must have dimension of either 2 or 3")
    if distribution:
        logits_t = torch.nn.functional.log_softmax(logits_t, 2)
    neg_inf = torch.tensor(-float("inf"), device=logits_t.device)
    batch_size, old_width, num_classes = logits_t.shape
    if log_prior is None:
        log_prior = torch.full(
            (batch_size, old_width),
            -torch.log(torch.tensor(float(num_classes))),
            dtype=logits_t.dtype,
            device=logits_t.device,
        )
    elif tuple(log_prior.shape) == (batch_size,) and old_width == 1:
        log_prior = log_prior.unsqueeze(1)
    elif log_prior.shape != logits_t.shape[:-1]:
        raise RuntimeError(
            "If logits_t of shape {} then log_prior must have shape {}".format(
                (batch_size, old_width, num_classes), (batch_size, old_width),
            )
        )
    if prevent_eos and 0 <= eos < num_classes:
        # we have to put this before the num_done check so that it'll be
        # overwritten for paths that have finished already
        logits_t[..., eos] = neg_inf
    eos_set = None
    if y_prev is not None:
        if y_prev.dim() == 2:
            y_prev = y_prev.unsqueeze(2)
        if y_prev.shape[1:] != log_prior.shape:
            raise RuntimeError(
                "If logits_t of shape {} then y_prev must have shape "
                "(*, {}, {})".format(
                    (batch_size, old_width, num_classes), batch_size, old_width,
                )
            )
        eos_mask = y_prev[-1].eq(eos)
        num_done = eos_mask.long().sum(1)
        if num_done.sum().item():
            if old_width < width and torch.any(num_done == old_width):
                raise RuntimeError(
                    "New beam width ({}) is wider than old beam width "
                    "({}), but all paths are already done in one or more "
                    "batch elements".format(width, old_width)
                )
            # we're going to treat class 0 as the sentinel for eos (even if
            # eos is a legit class label)
            done_classes = torch.full_like(logits_t[0, 0], neg_inf)
            done_classes[0] = 0.0
            logits_t = torch.where(eos_mask.unsqueeze(2), done_classes, logits_t,)
            # If eos_mask looks like this (vertical batch, horizontal beam):
            #   1 0 0 1 0
            #   0 1 0 0 0
            #   0 0 0 0 0
            # then eos_set will be
            #    0 -1 -1  3 -1
            #   -1  1 -1 -1 -1
            #   -1 -1 -1 -1 -1
            # s might look like
            #    1 2 3 3
            #    2 2 4 1
            #    1 2 3 4
            # we'll compare a single value from a row of s to a matched row
            # of eos_set. Any match means the beam had finished already. The
            # mask on y will be
            #    0 0 1 1
            #    0 0 0 1
            #    0 0 0 0
            # pretty funky
            eos_set = torch.where(
                eos_mask,
                torch.arange(old_width, device=logits_t.device),
                torch.tensor(-1, device=logits_t.device).expand(old_width),
            )
        t = y_prev.shape[0] + 1
    else:
        t = 1
    len_mask = None
    if lens is not None:
        if lens.shape != logits_t.shape[:1]:
            raise RuntimeError("lens must be of shape ({},)".format(batch_size))
        len_mask = lens.lt(t)
        if torch.any(len_mask):
            if old_width < width:
                raise RuntimeError(
                    "New beam width ({}) is wider than old beam width "
                    "({}), but all paths are already done in one or more "
                    "batch elements".format(width, old_width)
                )
        else:
            len_mask = None
    joint = log_prior.unsqueeze(2) + logits_t
    score, idxs = torch.topk(joint.view(batch_size, -1), width, dim=1)
    s = idxs // num_classes
    y = (idxs % num_classes).unsqueeze(0)
    if eos_set is not None:
        y_mask = (s.unsqueeze(2) == eos_set.unsqueeze(1)).any(2)
        y = y.masked_fill(y_mask, eos)
    if len_mask is not None:
        score = torch.where(len_mask.unsqueeze(1), log_prior[..., :width], score,)
        y = y.masked_fill(len_mask.unsqueeze(1).unsqueeze(0), eos)
        s = torch.where(
            len_mask.unsqueeze(1), torch.arange(width, device=logits_t.device), s,
        )
    if y_prev is not None:
        y_prev = y_prev.gather(2, s.unsqueeze(0).expand(t - 1, batch_size, width))
        y = torch.cat([y_prev, y], 0)
    return score, y, s


def time_distributed_return(r, gamma, batch_first=False):
    r"""Accumulate future local rewards at every time step

    In `reinforcement learning
    <https://en.wikipedia.org/wiki/Reinforcement_learning>`__, the return is
    defined as the sum of discounted future rewards. This function calculates
    the return for a given time step :math:`t` as

    .. math::

        R_t = \sum_{t'=t} \gamma^(t' - t) r_{t'}

    Where :math:`r_{t'}` gives the (local) reward at time :math:`t'` and
    :math:`\gamma` is the discount factor. :math:`\gamma \in [0, 1)` implies
    convergence, but this is not enforced here

    Parameters
    ----------
    r : torch.FloatTensor
        A two-dimensional tensor of shape ``(steps, batch_size)`` (or
        ``(batch_size, steps)`` if `batch_first` is :obj:`True`) of local
        rewards. The :math:`t` dimension is the step dimension
    gamma : float
        The discount factor
    batch_first : bool, optional

    Returns
    -------
    `R` : torch.FloatTensor
        Of the same shape as `r`

    See Also
    --------
    :ref:`Gradient Estimators`
        Provides an example of reinforcement learning that uses this function
    """
    if r.dim() != 2:
        raise RuntimeError("r must be 2 dimensional")
    if not gamma:
        return r
    if batch_first:
        exp = torch.arange(r.shape[-1], device=r.device, dtype=r.dtype)
        discount = torch.pow(gamma, exp)
        discount = (discount.unsqueeze(1) / discount.unsqueeze(0)).tril()
        R = torch.matmul(r, discount)
    else:
        exp = torch.arange(r.shape[0], device=r.device, dtype=r.dtype)
        discount = torch.pow(gamma, exp)
        discount = (discount.unsqueeze(0) / discount.unsqueeze(1)).triu()
        R = torch.matmul(discount, r)
    return R


def error_rate(
    ref,
    hyp,
    eos=None,
    include_eos=False,
    norm=True,
    batch_first=False,
    ins_cost=1.0,
    del_cost=1.0,
    sub_cost=1.0,
    warn=True,
):
    """Calculate error rates over a batch

    An error rate is merely a `Levenshtein (edit) distance
    <https://en.wikipedia.org/wiki/Levenshtein_distance>`__ normalized over
    reference sequence lengths.

    Given a reference (gold-standard) transcript long tensor `ref` of size
    ``(max_ref_steps, batch_size)`` if `batch_first` is :obj:`False` or
    ``(batch_size, max_ref_steps)`` otherwise, and a long tensor `hyp` of shape
    ``(max_hyp_steps, batch_size)`` or ``(batch_size, max_hyp_steps)``, this
    function produces a tensor `er` of shape ``(batch_size,)`` storing the
    associated error rates.

    `er` will not have a gradient, and is thus not directly suited to being a
    loss function

    Parameters
    ----------
    ref : torch.LongTensor
    hyp : torch.LongTensor
    eos : int, optional
        A special token in `ref` and `hyp` whose first occurrence in each
        batch indicates the end of a transcript. This allows for
        variable-length transcripts in the batch
    include_eos : bool, optional
        Whether to include the first instance of `eos` found in both `ref` and
        `hyp` as valid tokens to be computed as part of the distance. This is
        useful when gauging if a model is learning to emit the `eos` properly,
        but is not usually included in an evaluation. Only the first `eos` per
        transcript is included
    norm : bool, optional
        If :obj:`False`, will return edit distances instead of error rates
    batch_first : bool, optional
    ins_cost : float, optional
        The cost of an adding a superfluous token to a transcript in `hyp`
    del_cost : float, optional
        The cost of missing a token from `ref`
    sub_cost : float, optional
        The cost of swapping a token from `ref` with one from `hyp`
    warn : bool, optional
        Whether to display warnings on irregularities. Currently, this can
        happen in two ways:

        1. If :obj:`True` and `norm` is :obj:`True`, will warn when a reference
           transcription has zero length
        2. If `eos` is set and `include_eos` is :obj:`True`, will warn when a
           transcript does not include an `eos` symbol

    Returns
    -------
    er : torch.FloatTensor
    """
    er = _levenshtein(
        ref,
        hyp,
        eos,
        include_eos,
        batch_first,
        ins_cost,
        del_cost,
        sub_cost,
        warn,
        norm=norm,
    )
    return er


def optimal_completion(
    ref,
    hyp,
    eos=None,
    include_eos=True,
    batch_first=False,
    ins_cost=1.0,
    del_cost=1.0,
    sub_cost=1.0,
    padding=pydrobert.torch.INDEX_PAD_VALUE,
    exclude_last=False,
    warn=True,
):
    r"""Return a mask of next tokens of a minimum edit distance prefix

    Given a reference transcript `ref` of shape ``(max_ref_steps, batch_size)``
    (or ``(batch_size, max_ref_steps)`` if `batch_first` is :obj:`True`) and a
    hypothesis transcript `hyp` of shape ``(max_hyp_steps, batch_size)`` (or
    ``(batch_size, max_hyp_steps)``), this function produces a long tensor
    `optimals` of shape ``(max_hyp_steps + 1, batch_size, max_unique_next)``
    (or ``(batch_size, max_hyp_steps + 1, max_unique_next)``), where
    ``max_unique_next <= max_ref_steps``, of the unique tokens that could be
    added to the hypothesis prefix ``hyp[:prefix_len, batch]`` such that some
    remaining suffix concatenated to the prefix would result in a minimal edit
    distance. See below for an example

    Parameters
    ----------
    ref : torch.LongTensor
    hyp : torch.LongTensor
    eos : int, optional
        A special token in `ref` and `hyp` whose first occurrence in each
        batch indicates the end of a transcript. This allows for
        variable-length transcripts in the batch
    include_eos : bool, optional
        Whether to include the first instance of `eos` found in both `ref` and
        `hyp` as valid tokens to be computed as part of the distance and next
        tokens for a suffix. Only the first `eos` per transcript is included
    batch_first : bool, optional
    ins_cost : float, optional
        The cost of an adding a superfluous token to a transcript in `hyp`
    del_cost : float, optional
        The cost of missing a token from `ref`
    sub_cost : float, optional
        The cost of swapping a token from `ref` with one from `hyp`
    padding : int, optional
        The value to right-pad unequal-length sequences with
    exclude_last : bool, optional
        If true, will exclude the final prefix, consisting of the entire
        transcript, from the returned `optimals`. Optimals will be of shape
        ``(max_hyp_steps, batch_size, max_unique_next)``
    warn : bool, optional
        Whether to display warnings on irregularities. Currently, this only
        occurs when `eos` is set, `include_eos` is :obj:`True`, and a
        transcript does not contain the `eos` symbol

    Returns
    -------
    optimals : torch.LongTensor

    Examples
    --------

    Consider the reference text "foot" and the hypothesis text "bot". The below
    shows the matrix used to calculate edit distances between them::

        \ _ f o o t
        _ 0 1 2 3 4
        b 1 1 2 3 4
        o 2 2 1 2 3
        t 3 3 2 2 2

    If ``prefix_len == 0``, then the prefix is "", and "f" (from the suffix
    "foot") is the only subsequent token that would not increase the edit
    distance from that of the prefix (0). If ``prefix_len == 1``, then the
    prefix is "b". To arrive at the minimum edit distance for "b", one either
    treats "b" as an insertion or a substitution for "f", yielding suffixes
    "foot" and "oot". Thus, the subsequent token could be "f" or "o". For the
    prefix "bo", the minimum edit distance is achieved by first substituting
    "f" for "b", then substituting "o" for "o", resulting in the suffix "ot"
    and the next optimal character "o". Finally, for ``prefix_len == 3`` and
    prefix "bot", there are many operations that can produce the minimum edit
    distance of 2, resulting in one of the suffixes "ot", "t", and "". The
    latter suffix requires no more tokens and so any operation would increase
    the edit distance. Thus the optimal next tokens could be "o" or "t".

    Plugging "foot" and "bot" into this function, we get the prefixes:

    >>> ref_text, hyp_text = "foot", "bot"
    >>> ref = torch.tensor([ord(c) for c in ref_text]).unsqueeze(1)
    >>> hyp = torch.tensor([ord(c) for c in hyp_text]).unsqueeze(1)
    >>> optimal = optimal_completion(ref, hyp).squeeze(1)
    >>> for prefix_len, o_for_pr in enumerate(optimal):
    ...     o_for_pr = o_for_pr.masked_select(o_for_pr.ge(0)).tolist()
    ...     print('prefix={}: {}'.format(
    ...         hyp_text[:prefix_len], ','.join([chr(i) for i in o_for_pr])))
    prefix=: f
    prefix=b: f,o
    prefix=bo: o
    prefix=bot: o,t

    See Also
    --------
    pydrobert.torch.layers.HardOptimalCompletionDistillationLoss
        A loss function that uses these optimal completions to train a model
    """
    mask = _levenshtein(
        ref,
        hyp,
        eos,
        include_eos,
        batch_first,
        ins_cost,
        del_cost,
        sub_cost,
        warn,
        return_mask=True,
        exclude_last=exclude_last,
    )
    max_hyp_steps_p1, max_ref_steps, batch_size = mask.shape
    targets = []
    if batch_first:
        for mask_bt, ref_bt in zip(mask.transpose(0, 2), ref):
            for mask_bt_hyp in mask_bt.t():
                targets.append(torch.unique(ref_bt.masked_select(mask_bt_hyp)))
    else:
        for mask_hyp in mask:
            for mask_hyp_bt, ref_bt in zip(mask_hyp.t(), ref.t()):
                targets.append(torch.unique(ref_bt.masked_select(mask_hyp_bt)))
    targets = torch.nn.utils.rnn.pad_sequence(
        targets, padding_value=padding, batch_first=True
    )
    if batch_first:
        targets = targets.view(batch_size, max_hyp_steps_p1, -1)
    else:
        targets = targets.view(max_hyp_steps_p1, batch_size, -1)
    return targets


def prefix_error_rates(
    ref,
    hyp,
    eos=None,
    include_eos=True,
    norm=True,
    batch_first=False,
    ins_cost=1.0,
    del_cost=1.0,
    sub_cost=1.0,
    padding=pydrobert.torch.INDEX_PAD_VALUE,
    exclude_last=False,
    warn=True,
):
    """Compute the error rate between ref and each prefix of hyp

    Given a reference transcript `ref` of shape ``(max_ref_steps, batch_size)``
    (or ``(batch_size, max_ref_steps)`` if `batch_first` is :obj:`True`) and a
    hypothesis transcript `hyp` of shape ``(max_hyp_steps, batch_size)`` (or
    ``(batch_size, max_hyp_steps)``), this function produces a tensor
    `prefix_ers` of shape ``(max_hyp_steps + 1, batch_size)`` (or
    ``(batch_size, max_hyp_steps + 1))`` which contains the error rates for
    each prefix of each hypothesis, starting from the empty prefix

    Parameters
    ----------
    ref : torch.LongTensor
    hyp : torch.LongTensor
    eos : int, optional
        A special token in `ref` and `hyp` whose first occurrence in each
        batch indicates the end of a transcript. This allows for
        variable-length transcripts in the batch
    include_eos : bool, optional
        Whether to include the first instance of `eos` found in both `ref` and
        `hyp` as valid tokens to be computed as part of the distance.
        Only the first `eos` per transcript is included
    norm : bool, optional
        If :obj:`False`, will return edit distances instead of error rates
    batch_first : bool, optional
    ins_cost : float, optional
        The cost of an adding a superfluous token to a transcript in `hyp`
    del_cost : float, optional
        The cost of missing a token from `ref`
    sub_cost : float, optional
        The cost of swapping a token from `ref` with one from `hyp`
    padding : int, optional
        The value to right-pad the error rates of unequal-length sequences with
        in `prefix_ers`
    exclude_last : bool, optional
        If true, will exclude the final prefix, consisting of the entire
        transcript, from the returned `dists`. `dists` will be of shape
        ``(max_hyp_steps, batch_size, max_unique_next)``
    warn : bool, optional
        Whether to display warnings on irregularities. Currently, this only
        occurs when `eos` is set, `include_eos` is :obj:`True`, and a
        transcript does not contain the `eos` symbol

    Returns
    -------
    prefix_ers : torch.tensor

    See Also
    --------
    :ref:`Gradient Estimators`
        Provides an example where this function is used to determine a reward
        function for reinforcement learning
    """
    prefix_ers = _levenshtein(
        ref,
        hyp,
        eos,
        include_eos,
        batch_first,
        ins_cost,
        del_cost,
        sub_cost,
        warn,
        norm=norm,
        return_prf_dsts=True,
        exclude_last=exclude_last,
        padding=padding,
    )
    return prefix_ers


def random_walk_advance(
    logits_t,
    num_samp,
    y_prev=None,
    eos=pydrobert.torch.INDEX_PAD_VALUE,
    lens=None,
    prevent_eos=False,
    include_relaxation=False,
):
    r"""Advance a random walk of sequences

    Suppose a model outputs a un-normalized log-probability distribution over
    the next element of a sequence in `logits_t` s.t.

    .. math::

        Pr(y_t = c) = exp(logits_{t,c}) / \sum_k exp(logits_k)

    We assume :math:`logits_t` is a function of what comes before
    :math:`logits_t = f(logits_{<t}, y_{<t})`. Alternatively, letting
    :math:`s_t = (logits_t, y_t)`, :math:`s` is a Markov Chain. A model is
    auto-regressive if :math:`f` depends on :math:`y_{<t}`, and is not
    auto-regressive if :math:`logits_t = f(logits_{<t})`.

    A random walk can be performed over a Markov Chain by sampling the elements
    :math:`y_t` of the greater sequence `y` one at a time, according to
    :math:`Pr(y_t = c)`. This allows us to sample the distribution
    :math:`Pr(Y)`.

    This function is called at every time step. It updates the sequences
    being built (`y_prev`) with one additional token and returns `y`. This
    function is intended to be coupled with an auto-regressive model, where
    `logits_t` is not known until :math:`y_t` is known. If the model is
    not auto-regressive, it is much more efficient to gather all `logits_t`
    into one :math:`logits` and sample all at once. See the examples section
    below for both behaviours

    Parameters
    ----------
    logits_t : torch.FloatTensor
        The conditional probabilities over class labels for the current time
        step. Either of shape ``(batch_size, old_samp, num_classes)``,
        where ``old_samp`` is the number of samples in the previous time
        step, or ``(batch_size, num_classes)``, where it is assumed that
        ``old_samp == 1``
    num_samp : int
        The number of samples to be drawn. Either ``old_samp == 1`` and/or
        ``num_samp <= old_samp`` must be :obj:`True`. That is, either all
        samples will share the same prefix, or we are building off a subset of
        the samples from ``y_prev`` (in this case, always the first `num_samp`)
    y_prev : torch.LongTensor, optional
        A tensor of shape ``(t - 1, batch_size, old_samp)`` or
        ``(t - 1, batch_size)`` specifying :math:`y_{<t}`. If unspecified,
        it is assumed that ``t == 1``
    eos : int, optional
        A special end-of-sequence symbol indicating that the beam has ended.
        Can be a class index. If this value occurs in in
        ``y_prev[-1, bt, smp]`` for some batch ``bt`` and sample ``smp``,
        `eos` will be appended to ``y_prev[:, bt, smp]``
    lens : torch.LongTensor, optional
        A tensor of shape ``(batch_size,)``. If ``t > lens[bt]`` for some
        batch ``bt``, all samples for ``bt`` will be considered finished. `eos`
        will be appended to `y_prev`
    prevent_eos : bool, optional
        Setting this flag to :obj:`True` will keep `eos` targets from being
        drawn unless a sample has finished (either with a prior `eos` or
        through `lens`). Note that this will only have an effect when ``0 <=
        eos <= num_classes``
    include_relaxation : bool, optional
        If :obj:`True`, a tuple will be returned whose second element is `z`,
        see below

    Returns
    -------
    y : torch.LongTensor
        A long tensor of shape ``(t, batch_size, num_samp)`` of the sampled
        sequences so far. Note that, since :math:`y_t` are drawn `i.i.d.`,
        there is no guarantee of the uniqueness of each `num_samp` samples
    z : torch.FloatTensor
        Only included if `include_relaxation` is :obj:`True`. `z` is a sample
        of a continuous relaxation of the categorical distribution of `logits`
        of shape ``(batch_size, num_samp, num_classes). Assuming ``y_prev[-1,
        bt, smp] != eos``, ``y[-1, bt, smp] == z[bt, smp].argmax(dim-1)``. If
        ``y_prev[-1, bt, smp] == eos``, ``z[bt, smp, :] = -infinity``. The
        primary purpose of `z` is to be used as an argument (alongside `y`) in
        more complicated gradient estimators from
        :mod:`pydrobert.torch.estimators`

    Examples
    --------

    Here is an example of random path sampling with a non-auto-regressive
    RNN. It does not need this function, and can take advantage of packed
    sequences for efficiency and gradient validity.

    >>> N, I, C, T, W, H, eos = 5, 4, 10, 100, 6, 15, 0
    >>> rnn = torch.nn.RNN(I, H)
    >>> ff = torch.nn.Linear(H, C)
    >>> inp = torch.rand(T, N, I)
    >>> lens = torch.randint(1, T + 1, (N,)).sort(descending=True)[0]
    >>> packed_inp = torch.nn.utils.rnn.pack_padded_sequence(inp, lens)
    >>> packed_h, _ = rnn(packed_inp)
    >>> packed_logits = ff(packed_h[0])
    >>> packed_logits_dup = packed_logits.detach().unsqueeze(1)
    >>> packed_logits_dup = packed_logits_dup.expand(-1, W, -1)  # (flat, W, C)
    >>> packed_y = torch.distributions.Categorical(
    ...     logits=packed_logits_dup).sample()  # (flat, W)
    >>> # we pad y with "eos" to ensure each sample is done by its length,
    >>> # but "eos" may have occurred beforehand
    >>> y = torch.nn.utils.rnn.pad_packed_sequence(
    ...     torch.nn.utils.rnn.PackedSequence(
    ...         packed_y, batch_sizes=packed_h[1]),
    ...     padding_value=eos, total_length=T,
    ... )[0]  # (T, N, W) (batch index gets inserted as 2nd dim)

    Here is an auto-regressive RNN that uses this function to build partial
    samples into `y`

    >>> N, I, C, T, W, H, eos, start = 5, 5, 10, 100, 5, 10, 0, -1
    >>> cell = torch.nn.RNNCell(I + 1, H)
    >>> ff = torch.nn.Linear(H, C)
    >>> inp = torch.rand(T, N, I)
    >>> y = torch.full((1, N, 1), start, dtype=torch.long)
    >>> h_t = torch.zeros(N, 1, H)
    >>> for inp_t in inp:
    >>>     y_tm1 = y[-1]
    >>>     old_samp = y_tm1.shape[-1]
    >>>     inp_t = inp_t.unsqueeze(1).expand(N, old_samp, I)
    >>>     x_t = torch.cat([inp_t, y_tm1.unsqueeze(2).float()], -1)
    >>>     h_t = cell(
    ...         x_t.view(N * old_samp, I + 1),
    ...         h_t.view(N * old_samp, H),
    ...     ).view(N, old_samp, H)
    >>>     logits_t = ff(h_t)
    >>>     y = random_walk_advance(logits_t, W, y, eos)
    >>>     if old_samp == 1:
    >>>         h_t = h_t.expand(-1, W, H).contiguous()

    Notes
    -----

    Unlike in the beam search, `logits_t` must be transformed into a
    probability distribution. Otherwise, we would not be able to sample the
    next step

    See Also
    --------
    :ref:`Gradient Estimators`
        Includes a use case for `include_relaxation`
    """
    if logits_t.dim() == 2:
        logits_t = logits_t.unsqueeze(1)
    elif logits_t.dim() != 3:
        raise RuntimeError("logits_t must have dimension of either 2 or 3")
    batch_size, old_samp, num_classes = logits_t.shape
    if prevent_eos and 0 <= eos < num_classes:
        logits_t[..., eos] = torch.tensor(-float("inf"), device=logits_t.device)
    if old_samp != 1 and num_samp > old_samp:
        raise RuntimeError("either old_samp == 1 or num_samp <= old_samp must be true")
    eos_mask = None
    if y_prev is not None:
        if y_prev.dim() == 2:
            y_prev = y_prev.unsqueeze(2)
        if y_prev.shape[1:] != logits_t.shape[:-1]:
            raise RuntimeError(
                "If logits_t of shape {} then y_prev must have shape "
                "(*, {}, {})".format(
                    (batch_size, old_samp, num_classes), batch_size, old_samp,
                )
            )
        y_prev = y_prev.expand(-1, -1, num_samp)
        eos_mask = y_prev[-1].eq(eos)
        if eos_mask.any():
            eos_mask = eos_mask[..., :num_samp]
        else:
            eos_mask = None
        t = y_prev.shape[0] + 1
    else:
        t = 1
    logits_t = logits_t.expand(-1, num_samp, -1)
    if lens is not None:
        if lens.shape != logits_t.shape[:1]:
            raise RuntimeError("lens must be of shape ({},)".format(batch_size))
        len_mask = lens.lt(t)
        if torch.any(len_mask):
            len_mask = len_mask.unsqueeze(1).expand(-1, num_samp)
            eos_mask = len_mask if eos_mask is None else (eos_mask | len_mask)
    u = torch.distributions.utils.clamp_probs(torch.rand_like(logits_t))
    log_theta = torch.nn.functional.log_softmax(logits_t, dim=-1)
    z = log_theta - torch.log(-torch.log(u))
    y = z.argmax(dim=-1)
    if eos_mask is not None:
        y = y.masked_fill(eos_mask, eos)
        z = z.masked_fill(eos_mask.unsqueeze(-1), -float("inf"))
    y = y.unsqueeze(0)
    if y_prev is not None:
        y = torch.cat([y_prev, y], 0)
    if include_relaxation:
        return y, z
    else:
        return y


def sequence_log_probs(logits, hyp, dim=0, eos=None):
    r"""Calculate joint log probability of sequences

    `logits` is a tensor of shape ``(..., steps, ..., num_classes)`` where
    ``steps`` enumerates the time/step `dim` -th dimension. `hyp` is a long
    tensor of shape ``(..., steps, ...)`` matching the shape of `logits` minus
    the last dimension. Letting :math:`t` index the step dimension and
    :math:`b` index all other shared dimensions of `logits` and `hyp`, this
    function outputs a tensor `log_probs` of the log-joint probability of
    sequences in the batch:

    .. math::

        \log Pr(samp_b = hyp_b) = \log \left(
            \prod_t Pr(samp_{b,t} == hyp_{b,t}; logits_{b,t})\right)

    :math:`logits_{b,t}` (with the last dimension free) characterizes a
    categorical distribution over ``num_classes`` tokens via a softmax
    function. We assume :math:`samp_{b,t}` is independent of
    :math:`samp_{b',t'}` given :math:`logits_t`.

    The resulting tensor `log_probs` is matches the shape of `logits` or
    `hyp` without the ``step`` and ``num_classes`` dimensions.

    This function can handle variable-length sequences in three ways:

    1. Any values of `hyp` not in ``[0, num_classes)`` will be considered
       padding and ignored.
    2. If `eos` (end-of-sentence) is set, the first occurrence at :math:`b,t`
       is included in the sequence, but all :math:`b,>t` are ignored. This is
       in addition to 1.
    3. `logits` and `hyp` may be :class:`torch.nn.utils.rnn.PackedSequence`
       objects. In this case, the packed sequence dimension is assumed to
       index ``steps`` (`dim` is ignored). The remaining batch dimension will
       always be stacked into dimension 0. This is also in addition to 1.

    Parameters
    ----------
    logits : torch.FloatTensor or torch.nn.utils.rnn.PackedSequence
    hyp : torch.LongTensor or torch.nn.utils.rnn.PackedSequence
    dim : int, optional
    eos : int or :obj:`None`, optional

    Returns
    -------
    log_prob : torch.FloatTensor

    Notes
    -----
    `dim` is relative to ``hyp.shape``, not ``logits.shape``

    See Also
    --------
    pydrobert.torch.layers.MinimumErrorRateLoss
        An example training regime that uses this function
    """
    if isinstance(logits, torch.nn.utils.rnn.PackedSequence):
        return _sequence_log_probs_packed(logits, hyp)
    if isinstance(hyp, torch.nn.utils.rnn.PackedSequence):
        raise RuntimeError("both hyp and logits must be packed sequences, or neither")
    if logits.shape[:-1] != hyp.shape:
        raise RuntimeError(
            "logits and hyp must have same shape (minus last dimension of " "logits)"
        )
    hyp_dim = hyp.dim()
    if dim < -hyp_dim or dim > hyp_dim - 1:
        raise RuntimeError(
            "Dimension out of range (expected to be in range of [{}, {}], but "
            "got {})".format(-hyp_dim, hyp_dim - 1, dim)
        )
    dim = (hyp_dim + dim) % hyp_dim
    steps = hyp.shape[dim]
    num_classes = logits.shape[-1]
    logits = torch.nn.functional.log_softmax(logits, -1)
    mask = hyp.lt(0) | hyp.ge(num_classes)
    if eos is not None:
        hyp_lens = _lens_from_eos(hyp, eos, dim) + 1
        if dim:
            hyp_lens = hyp_lens.unsqueeze(dim)
            if dim == hyp_dim - 1:
                hyp_lens = hyp_lens.unsqueeze(-1)
            else:
                hyp_lens = hyp_lens.flatten(dim + 1)
        else:
            hyp_lens = hyp_lens.view(1, -1)
        len_mask = torch.arange(steps, device=logits.device).unsqueeze(-1)
        len_mask = len_mask >= hyp_lens
        len_mask = len_mask.view_as(mask)
        mask = mask | len_mask
    hyp = hyp.masked_fill(mask, 0)
    logits = logits.gather(-1, hyp.unsqueeze(-1)).squeeze(-1)
    logits = logits.masked_fill(mask, 0.0)
    return logits.sum(dim)


def polyharmonic_spline(
    train_points,
    train_values,
    query_points,
    order,
    regularization_weight=0.0,
    full_matrix=True,
):
    """Guess values at query points using a learned polyharmonic spline

    A spline estimates a function ``f : points -> values`` from a fixed number of
    training points/knots and the values of ``f`` at those points. It does that by
    solving a series of piecewise linear equations between knots such that the values at
    the knots match the given values (and some additional constraints depending on the
    spline).

    This function based on the `interpolate_spline
    <https://www.tensorflow.org/addons/api_docs/python/tfa/image/interpolate_spline>`__
    function from Tensorflow, which implements a `Polyharmonic Spline
    <https://en.wikipedia.org/wiki/Polyharmonic_spline>`__. For technical details,
    consult the TF documentation.

    Parameters
    ----------
    train_points : torch.Tensor
        A tensor of shape ``(N, T, I)`` representing the training points/knots for
        ``N`` different functions. ``N`` is the batch dimension, ``T`` is the number
        of training points, and ``I`` is the size of the vector input to ``f``.
    train_values : torch.FloatTensor
        A float tensor of shape ``(N, T, O)`` of ``f`` evaluated on `train_points`.
        ``O`` is the size of the output vector of ``f``.
    query_points : torch.Tensor
        A tensor of shape ``(N, Q, I)`` representing the points you wish to have
        estimates for. ``Q`` is the number of such points
    order : int
        Order of the spline (> 0). 1 = linear. 2 = thin plate spline.
    regularization_weight : float, optional
        Weight placed on the regularization term. See TF for more info.
    full_matrix : bool, optional
        Whether to solve linear equations via a full concatenated matrix or a block
        decomposition. Setting to :obj:`True` better matches TF and appears to slightly
        improve numerical accuracy at the cost of twice the run time and more memory
        usage.

    Throws
    ------
    RuntimeError
        This function can return a :class`RuntimeError` when no unique spline can be
        estimated. In general, the spline will require at least ``I+1`` non-degenerate
        points (linearly independent). See the Wikipedia entry on splnes for more info.

    Returns
    -------
    query_values : torch.FloatTensor
        A tensor of shape ``(N, Q, O)`` of the values estimated by the spline
    """
    train_points = train_points.float()
    query_points = query_points.float()

    w, v = _solve_interpolation(
        train_points, train_values, order, regularization_weight, full_matrix
    )

    query_values = _apply_interpolation(w, v, train_points, query_points, order)
    return query_values


def dense_image_warp(
    image, flow, indexing="hw", mode="bilinear", padding_mode="border"
):
    """Warp an input image with per-pixel flow vectors

    Given an `image` and a `flow` field, generates a new image `warped` such that

    ::
        warped[n, c, h, w] = image[n, c, h - flow[n, h, w, 0], w - flow[n, h, w, 1]]

    If the reference indices ``h - ...`` and ``w - ...`` are not integers, the value is
    interpolated from the neighboring pixel values.

    This reproduces the functionality of Tensorflow's `dense_image_warp
    <https://www.tensorflow.org/addons/api_docs/python/tfa/image/dense_image_warp>`__,
    except `image` is in ``NCHW`` order instead of ``NHWC`` order. It wraps
    `torch.nn.functional.grid_sample`.

    Warning
    -------
    `flow` is not an optical flow. Please consult the TF documentation for more details.

    Parameters
    ----------
    image : torch.FloatTensor
        A float tensor of shape ``(N, C, H, W)``, where ``N`` is the batch dimension,
        ``C`` is the channel dimension, ``H`` is the height dimension, and ``W`` is the
        width dimension.
    flow : torch.Tensor
        A tensor of shape ``(N, H, W, 2)``.
    indexing : {'hw', 'wh'}, optional
        If `indexing` is ``"hw"``, ``flow[..., 0] = h``, the height index, and
        ``flow[..., 1] = w`` is the width index. If ``"wh"``, ``flow[..., 0] = w``
        and ``flow[..., 1] = h``. The default in TF is ``"hw"``, whereas torch's
        `grid_sample` is ``"wh"``
    mode : {'bilinear', 'nearest'}
        The method of interpolation. Either use bilinear interpolation or the nearest
        pixel value. The TF default is ``"bilinear"``
    padding_mode : {"border", "zeros", "reflection"}
        Controls how points outside of the image boundaries are interpreted.
        ``"border"``: copy points at around the border of the image. ``"zero"``:
        use zero-valued pixels. ``"reflection"``: reflect pixels into the image starting
        from the boundaries.

    Returns
    -------
    warped : torch.FloatTensor
        The warped image of shape ``(N, C, H, W)``.
    """

    # from tfa.image.dense_image_warp
    # output[n, c, h, w] = image[n, c, h - flow[n, h, w, 0], w - flow[n, h, w, 1]]
    # outside of image uses border

    # from torch.nn.functional.grid_sample
    # output[n, c, h, w] = image[n, c, h, f(grid[n, h, w, 1], H),
    # f(grid[n, h, w, 0], W)]
    # where
    # f(x, X) = ((x + 1) * X - 1) / 2
    # therefore
    # output[n, c, h, w] = image[n, c, ((grid[n, h, w, 1] + 1) * H - 1) / 2,
    #                                  ((grid[n, h, w, 0] + 1) * W - 1) / 2]
    #
    # ((grid[n, h, w, 1] + 1) * H - 1) / 2 = h - flow[n, h, w, 0]
    # grid[n, h, w, 1] = (2 * h - 2 * flow[n, h, w, 0] + 1) / H - 1
    # likewise
    # grid[n, h, w, 0] = (2 * w - 2 * flow[n, h, w, 1] + 1) / W - 1

    flow = flow.float()

    N, C, H, W = image.shape
    h = torch.arange(H, dtype=image.dtype, device=image.device)  # (H,)
    w = torch.arange(W, dtype=image.dtype, device=image.device)  # (W,)
    h, w = torch.meshgrid(h, w)  # (H, W), (H, W)
    if indexing == "hw":
        # grid_sample uses wh sampling, so we flip both the flow and hw along final axis
        hw = torch.stack((w, h), 2).unsqueeze(0)  # (1, H, W, 2)
        flow = flow.flip(-1)
    elif indexing == "wh":
        hw = torch.stack((w, h), 2).unsqueeze(0)  # (1, H, W, 2)
    else:
        raise ValueError("Invalid indexing! must be one of 'wh' or 'hw'")
    HW = torch.tensor([[[[W, H]]]], dtype=image.dtype, device=image.device)  # (1,1,1,2)
    grid = (2 * hw - 2 * flow + 1.0) / HW - 1.0

    return torch.nn.functional.grid_sample(
        image, grid, mode=mode, padding_mode=padding_mode, align_corners=False
    )


def sparse_image_warp(
    image,
    source_points,
    dest_points,
    indexing="hw",
    field_interpolation_order=2,
    field_regularization_weight=0.0,
    field_full_matrix=True,
    pinned_boundary_points=0,
    dense_interpolation_mode="bilinear",
    dense_padding_mode="border",
    include_flow=True,
):
    r"""Warp an image by specifying mappings between few control points

    Given a source image `image`, a few source coordinates `source_points` and their
    corresponding positions `dest_points` in the warped image, this function
    interpolates the remainder of the map with a polyharmonic spline and produces a
    warped image `warped`.

    This function mirrors the behaviour of Tensorflow's `sparse_image_warp
    <https://www.tensorflow.org/addons/api_docs/python/tfa/image/sparse_image_warp>`__,
    except `image` is in ``NCHW`` order instead of ``NHWC`` order. For more details,
    please consult their documentation.

    Parameters
    ----------
    image : torch.FloatTensor
        A float tensor of shape ``(N, C, H, W)``, where ``N`` is the batch dimension,
        ``C`` the channel dimension, ``H`` the image height, and ``W`` the image width.
    source_points : torch.Tensor
        A tensor of shape ``(N, M, 2)``, where ``M`` is the number of control points
        and the final dimension stores the coordinates of the control point in `image`.
    dest_points : torch.Tensor
        A tensor of shape ``(N, M, 2)`` such that the point ``source_points[n, m, :]``
        in `image` will be mapped to ``dest_points[n, m, :]`` in `warped`.
    indexing : {'hw', 'wh'}, optional
        If `indexing` is ``"hw"``, ``source_points[n, m, 0]`` and
        ``dest_points[n, m, 0]`` index the height dimension in `image` and `warped`,
        respectively, and ``source_points[n, m, 1]`` and ``dest_points[n, m, 1]`` the
        width dimension. If `indexing` is ``"wh"``, the width dimension is the 0-index
        and height the 1.
    field_interpolation_order : int, optional
        The order of the polyharmonic spline used to interpolate the rest of the points
        from the control. See :func:`polyharmonic_spline` for more info.
    field_regularization_weight : int, optional
        The regularization weight of the polyharmonic spline used to interpolate the
        rest of the points from the control. See :func:`polyharmonic_spline` for more
        info.
    field_full_matrix : bool, optional
        Determines the method of calculating the polyharmonic spline used to interpolate
        the rest of the points from the control. See :func:`polyharmonic_spline` for
        more info.
    pinned_boundary_points : int, optional
        Dictates whether and how many points along the boundary of `image` are mapped
        identically to points in `warped`. This keeps the boundary of the `image` from
        being pulled into the interior of `warped`. When :obj:`0`, no points are added.
        When :obj:`1`, four points are added, one in each corner of the image. When
        ``k > 2``, one point in each corner of the image is added, then ``k - 1``
        equidistant points along each of the four edges, totaling ``4 * k`` points.
    dense_interpolation_mode : {'bilinear', 'nearest'}, optional
        The method with which partial indices in the derived mapping are interpolated.
        See :func:`dense_image_warp` for more info.
    dense_padding_mode : {'border', 'zero', 'reflection'}, optional
        What to do when points in the derived mapping fall outside of the boundaries.
        See :func:`dense_image_warp` for more info.
    include_flow : bool, optional
        If :obj:`True`, include the flow field `flow` interpolated from the control
        points in the return value.

    Returns
    -------
    warped[, flow] : torch.FloatTensor[, torch.FloatTensor]
        `warped` is a float tensor of shape ``(N, C, H, W)`` containing the warped
        images. If `include_flow` is :obj:`True`, `flow`, a float tensor of shape
        ``(N, H, W, 2)``. ``flow[n, h, w, :]`` is the flow for coordinates ``h, w``
        in whatever order was specified by `indexing`. See :func:`dense_image_warp`
        for more details.
    """

    # all our computations assume "wh" ordering, so we flip it here if necessary.
    # Though unintuitive, we need this for our call to grid_sample
    if indexing == "hw":
        source_points = source_points.flip(-1)
        dest_points = dest_points.flip(-1)

    source_points = source_points.float()
    dest_points = dest_points.float()

    N, C, H, W = image.shape
    WH = torch.tensor([[W, H]] * N, dtype=image.dtype, device=image.device)

    M = source_points.shape[1]
    if not M:
        return image

    if pinned_boundary_points > 0:
        pinned_points = _deterimine_pinned_points(pinned_boundary_points, WH)
        source_points = torch.cat([source_points, pinned_points], 1)  # (N,M',2)
        dest_points = torch.cat([dest_points, pinned_points], 1)  # (N,M+4k=M',2)
        # now just pretend M' was M all along

    H_range = torch.arange(H, dtype=image.dtype, device=image.device)  # (H,)
    W_range = torch.arange(W, dtype=image.dtype, device=image.device)  # (W,)
    h, w = torch.meshgrid(H_range, W_range)  # (H, W), (H, W)
    query_points = torch.stack([w.flatten(), h.flatten()], 1)  # (H * W, 2)

    if include_flow:
        train_points = dest_points
        train_values = dest_points - source_points
        flow = polyharmonic_spline(
            train_points,
            train_values,
            query_points.unsqueeze(0).expand(N, H * W, 2),
            field_interpolation_order,
            regularization_weight=field_regularization_weight,
            full_matrix=field_full_matrix,
        )

        flow = flow.view(N, H, W, 2)

        warped = dense_image_warp(
            image,
            flow,
            indexing="wh",
            mode=dense_interpolation_mode,
            padding_mode=dense_padding_mode,
        )

        if indexing == "hw":
            flow = flow.flip(-1)

        return warped, flow
    else:
        # If we can return just the warped image, we can bypass our call to
        # dense_image_warp by interpolating the 'grid' parameter of 'grid_sample'
        # instead of the 'flow' parameter of 'dense_image_warp'
        # coord = ((grid + 1) * size - 1) / 2
        # grid = (2 coord + 1) / size - 1
        train_points = dest_points  # (N, M, 2)
        train_values = (2.0 * source_points + 1.0) / WH.unsqueeze(1) - 1.0  # (N, M, 2)

        grid = polyharmonic_spline(
            train_points,
            train_values,
            query_points.unsqueeze(0).expand(N, H * W, 2),
            field_interpolation_order,
            regularization_weight=field_regularization_weight,
            full_matrix=field_full_matrix,
        )

        grid = grid.view(N, H, W, 2)

        warped = torch.nn.functional.grid_sample(
            image,
            grid,
            mode=dense_interpolation_mode,
            padding_mode=dense_padding_mode,
            align_corners=False,
        )

        return warped


def _deterimine_pinned_points(k, sizes):

    w_max = (sizes[:, :1] - 1).expand(-1, k + 1)  # (N, k+1)
    h_max = (sizes[:, 1:] - 1).expand(-1, k + 1)  # (N, k+1)
    range_ = torch.linspace(
        0.0, 1.0, k + 1, dtype=sizes.dtype, device=sizes.device
    )  # (k+1,)
    w_range = w_max * range_  # (N, k+1)
    h_range = h_max * range_  # (N, k+1)
    zeros = torch.zeros_like(w_range)  # (N, k+1)

    # (0, 0) -> (W - 1, 0) inclusive
    bottom_edge = torch.stack([w_range, zeros], 2)  # (N, k+1, 2)
    # (0, 0) -> (0, H - 1) exclusive
    left_edge = torch.stack([zeros[:, 1:-1], h_range[:, 1:-1]], 2)  # (N, k-1, 2)
    # (0, H - 1) -> (W - 1, H - 1) inclusive
    top_edge = torch.stack([w_range, h_max], 2)  # (N, k+1, 2)
    # (W - 1, 0) -> (W - 1, H - 1) exclusive
    right_edge = torch.stack([w_max[:, 1:-1], h_range[:, 1:-1]], 2)  # (N, k-1, 2)

    return torch.cat([bottom_edge, left_edge, top_edge, right_edge], 1)  # (N, 4k, 2)


def _solve_interpolation(c, f, k, reg, full):
    # based on
    # https://mathematica.stackexchange.com/questions/65763/understanding-polyharmonic-splines
    # Symbol map (theirs => ours)
    # x,y => c  (N, T, I)
    # A => A    (N, T, T)
    # W => B    (N, T, I+1)
    # v => w    (N, T, O)
    # bb => v   (N, I+1, O)
    # wa => f   (N, T, O)
    r_cc = torch.cdist(c, c)  # (N, T, T)
    A = _phi(r_cc, k)  # (N, T, T)
    if reg > 0.0:
        A = A + torch.eye(A.shape[1], dtype=A.dtype, device=A.device).unsqueeze(0) * reg
    B = torch.cat([c, torch.ones_like(c[..., :1])], 2)  # (N, T, I+1)

    if full:
        # full matrix method (TF)
        ABt = torch.cat([A, B.transpose(1, 2)], 1)  # (N, T+I+1, T)
        zeros = torch.zeros(
            (B.shape[0], B.shape[2], B.shape[2]), device=B.device, dtype=B.dtype
        )
        B0 = torch.cat([B, zeros], 1,)  # (N, T+I+1, I+1)
        ABtB0 = torch.cat([ABt, B0], 2)  # (N, T+I+1, T+I+1)
        zeros = torch.zeros(
            (B.shape[0], B.shape[2], f.shape[2]), device=f.device, dtype=f.dtype
        )
        f0 = torch.cat([f, zeros], 1,)  # (N, T+I+1, O)
        wv, _ = torch.solve(f0, ABtB0)
        w, v = wv[:, : B.shape[1]], wv[:, B.shape[1] :]
    else:
        # block decomposition
        Ainv = torch.inverse(A)  # (N, T, T)
        Ainv_f = torch.bmm(Ainv, f)  # (N, T, O)
        Ainv_B = torch.bmm(Ainv, B)  # (N, T, I+1)
        Bt = B.transpose(1, 2)  # (N, I+1, T)
        Bt_Ainv_B = torch.bmm(Bt, Ainv_B)  # (N, I+1, I+1)
        Bt_Ainv_f = torch.bmm(Bt, Ainv_f)  # (N, I+1, O)
        v, _ = torch.solve(Bt_Ainv_f, Bt_Ainv_B)  # (N, I+1, O)
        Ainv_B_v = torch.bmm(Ainv_B, v)  # (N, T, O)
        w = Ainv_f - Ainv_B_v  # (N, T, O)

    # orthagonality constraints
    # assert torch.allclose(w.sum(1), torch.tensor(0.0, device=w.device)), w.sum()
    # assert torch.allclose(
    #     torch.bmm(w.transpose(1, 2), c), torch.tensor(0.0, device=w.device)
    # ), torch.bmm(w.transpose(1, 2), c).sum()

    return w, v


def _apply_interpolation(w, v, c, x, k):
    r = torch.cdist(x, c)  # (N, Q, T)
    phi_r = _phi(r, k)  # (N, Q, T)
    phi_r_w = torch.bmm(phi_r, w)  # (N, Q, O)
    x1 = torch.cat([x, torch.ones_like(x[..., :1])], 2)  # (N, Q, I+1)
    x1_v = torch.bmm(x1, v)  # (N, Q, O)
    return phi_r_w + x1_v


def _phi(r, k):
    if k % 2:
        return r ** k
    else:
        return (r ** k) * (torch.clamp(r, min=torch.finfo(r.dtype).eps)).log()


def _lens_from_eos(tok, eos, dim):
    # length to first eos (exclusive)
    mask = tok.eq(eos)
    x = torch.cumsum(mask, dim, dtype=torch.long)
    max_, argmax = (x.eq(1) & mask).max(dim)
    return argmax.masked_fill(max_.eq(0), tok.shape[dim])


def _sequence_log_probs_packed(logits, hyp):
    if not isinstance(hyp, torch.nn.utils.rnn.PackedSequence):
        raise RuntimeError("both hyp and logits must be packed sequences, or neither")
    logits, logits_lens = logits.data, logits.batch_sizes
    hyp, hyp_lens = hyp.data, hyp.batch_sizes
    if (hyp_lens != logits_lens).any():
        raise RuntimeError("hyp and logits must have the same sequence lengths")
    if logits.shape[:-1] != hyp.shape:
        raise RuntimeError(
            "logits and hyp must have same shape (minus last dimension of " "logits)"
        )
    num_classes = logits.shape[-1]
    logits = torch.nn.functional.log_softmax(logits, -1)
    not_class_mask = hyp.lt(0) | hyp.ge(num_classes)
    hyp = hyp.masked_fill(not_class_mask, 0)
    logits = logits.gather(-1, hyp.unsqueeze(-1)).squeeze(-1)
    logits = logits.masked_fill(not_class_mask, 0.0)
    logits_split = logits.split(logits_lens.tolist(), 0)
    return torch.stack(tuple(x.sum(0) for x in logits_split), 0)


def _levenshtein(
    ref,
    hyp,
    eos,
    include_eos,
    batch_first,
    ins_cost,
    del_cost,
    sub_cost,
    warn,
    norm=False,
    return_mask=False,
    return_prf_dsts=False,
    exclude_last=False,
    padding=None,
):
    assert not return_mask or not return_prf_dsts
    if ref.dim() != 2 or hyp.dim() != 2:
        raise RuntimeError("ref and hyp must be 2 dimensional")
    if batch_first:
        ref = ref.t()
        hyp = hyp.t()
    ref = ref.detach()
    hyp = hyp.detach()
    max_ref_steps, batch_size = ref.shape
    max_hyp_steps, batch_size_ = hyp.shape
    device = ref.device
    if batch_size != batch_size_:
        raise RuntimeError(
            "ref has batch size {}, but hyp has {}".format(batch_size, batch_size_)
        )
    if eos is not None:
        ref_lens = _lens_from_eos(ref, eos, 0)
        hyp_lens = _lens_from_eos(hyp, eos, 0)
        if include_eos:
            ref_eq_mask = ref_lens == max_ref_steps
            ref_lens = ref_lens + 1
            if ref_eq_mask.any():
                if warn:
                    warnings.warn(
                        "include_eos=True, but a transcription in ref did not "
                        "contain the eos symbol ({}). To suppress this "
                        "warning, set warn=False".format(eos)
                    )
                ref_lens = torch.where(ref_eq_mask, ref_lens - 1, ref_lens)
            hyp_eq_mask = hyp_lens == max_hyp_steps
            hyp_lens = hyp_lens + 1
            if hyp_eq_mask.any():
                if warn:
                    warnings.warn(
                        "include_eos=True, but a transcription in hyp did not "
                        "contain the eos symbol ({}). To suppress this "
                        "warning, set warn=False".format(eos)
                    )
                hyp_lens = torch.where(hyp_eq_mask, hyp_lens - 1, hyp_lens)
            del ref_eq_mask, hyp_eq_mask
    else:
        ref_lens = torch.full_like(ref[0], max_ref_steps)
        hyp_lens = torch.full_like(hyp[0], max_hyp_steps)
    ins_cost = torch.tensor(float(ins_cost), device=device)
    del_cost = torch.tensor(float(del_cost), device=device)
    sub_cost = torch.tensor(float(sub_cost), device=device)
    batch_range = torch.arange(batch_size, device=device)
    if return_mask:
        # this dtype business is a workaround for different default mask
        # types < 1.2.0 and > 1.2.0
        mask = torch.empty(
            (max_hyp_steps + (0 if exclude_last else 1), max_ref_steps, batch_size),
            device=device,
            dtype=ins_cost.eq(ins_cost).dtype,
        )
        mask[0, 0] = 1
        mask[0, 1:] = 0
    elif return_prf_dsts:
        assert padding is not None
        prefix_ers = torch.empty(
            (max_hyp_steps + (0 if exclude_last else 1), batch_size),
            device=device,
            dtype=torch.float,
        )
        prefix_ers[0] = ref_lens
    # direct row down corresponds to insertion
    # direct col right corresponds to a deletion
    # we vectorize as much as we can. Neither substitutions nor insertions
    # require values from the current row to be computed, and since the last
    # row can't be altered, we can easily vectorize there.
    # To vectorize deletions, we use del_matrix. It has entries
    #
    # 0   inf inf inf ...
    # d   0   inf inf ...
    # 2d  d   0   inf ...
    # ...
    #
    # Where "d" is del_cost. When we sum with the intermediate values of the
    # next row "v" (containing the minimum of insertion and subs costs), we get
    #
    # v[0]    inf     inf     inf ...
    # v[0]+d  v[1]    inf     inf ...
    # v[0]+2d v[1]+d  v[2]    inf ...
    # ...
    #
    # And we take the minimum of each row. The dynamic programming algorithm
    # for levenshtein would usually handle deletions as:
    #
    # for i=1..|v|:
    #     v[i] = min(v[i], v[i-1]+d)
    #
    # if we unroll the loop, we get the minimum of the elements of each row of
    # the above matrix
    row = torch.arange(max_ref_steps + 1, device=device, dtype=torch.float) * del_cost
    del_mat = row.unsqueeze(1) - row
    del_mat = del_mat + torch.full_like(del_mat, float("inf")).triu(1)
    del_mat = del_mat.unsqueeze(-1)  # batch
    row = row.unsqueeze(1).expand(max_ref_steps + 1, batch_size)
    last_row = torch.empty_like(row)
    for hyp_idx in range(1, max_hyp_steps + 1):
        last_row = row
        row = torch.where(hyp_lens < hyp_idx, last_row, last_row + ins_cost)
        sub_row = torch.where(
            ref == hyp[hyp_idx - 1], last_row[:-1], last_row[:-1] + sub_cost,
        )
        row[1:] = torch.min(row[1:], sub_row)
        row = (del_mat + row).min(1)[0]
        if return_mask and (hyp_idx < max_hyp_steps or not exclude_last):
            # As proven in the OCD paper, the optimal targets are always the
            # first character of a suffix of the reference transcript that
            # remains to be aligned. The levenshtein operation
            # corresponding to what we do with that target would be a matched
            # substitution (i.e. hyp's next token is the OCD target, resulting
            # in no change in cost from the prefix). Thus, given a levenshtein
            # matrix for one of these OCD targets (which is this matrix,
            # except for the final row), the minimal values on the final row
            # sit on a diagonal from the minimal values of the current row.
            mins = row.min(0, keepdim=True)[0]
            row_mask = row[:-1] == mins
            if exclude_last:
                row_mask = row_mask & (hyp_idx < hyp_lens)
            else:
                row_mask = row_mask & (hyp_idx <= hyp_lens)
            mask[hyp_idx] = row_mask
        elif return_prf_dsts and (hyp_idx < max_hyp_steps or not exclude_last):
            prefix_ers[hyp_idx] = row[ref_lens, batch_range]
    if return_mask:
        mask = mask & (
            (
                torch.arange(max_ref_steps, device=device)
                .unsqueeze(1)
                .expand(max_ref_steps, batch_size)
                < ref_lens
            ).unsqueeze(0)
        )
        return mask
    elif return_prf_dsts:
        if norm:
            prefix_ers = prefix_ers / ref_lens.float()
            zero_mask = ref_lens.eq(0).unsqueeze(0)
            if zero_mask.any():
                if warn:
                    warnings.warn(
                        "ref contains empty transcripts. Error rates will be "
                        "0 for prefixes of length 0, 1 otherwise. To suppress "
                        "this warning, set warn=False"
                    )
                prefix_ers = torch.where(
                    zero_mask,
                    (
                        torch.arange(prefix_ers.shape[0], device=device)
                        .gt(0)
                        .float()
                        .unsqueeze(1)
                        .expand_as(prefix_ers)
                    ),
                    prefix_ers,
                )
        prefix_ers = prefix_ers.masked_fill(
            (
                torch.arange(prefix_ers.shape[0], device=device)
                .unsqueeze(1)
                .ge(hyp_lens + (0 if exclude_last else 1))
            ),
            padding,
        )
        if batch_first:
            prefix_ers = prefix_ers.t()
        return prefix_ers
    er = row[ref_lens, batch_range]
    if norm:
        er = er / ref_lens.float()
        zero_mask = ref_lens.eq(0)
        if zero_mask.any():
            if warn:
                warnings.warn(
                    "ref contains empty transcripts. Error rates for entries "
                    "will be 1 if any insertion and 0 otherwise. To suppress "
                    "this warning, set warn=False"
                )
            er = torch.where(zero_mask, hyp_lens.gt(0).float(), er,)
    return er



# Copyright 2019 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2019 Sean Robertson"
__all__ = [
    'command_line',
    'data',
    'estimators',
    'INDEX_PAD_VALUE',
    'layers',
    'training',
    'util',
]


'''The value to pad index-based tensors with

Batched operations often involve variable-width input. This value is used to
right-pad indexed-based tensors with to indicate that this element should be
ignored.

The default value (-100) was chosen to coincide with the PyTorch 1.0 default
for ``ignore_index`` in the likelihood losses
'''
INDEX_PAD_VALUE = -100
